"""Smart file handeling"""

import uuid
from typing import BinaryIO, Optional

import smart_open

from plynx.utils.config import StorageConfig, get_storage_config

_GLOBAL_STORAGE_CONFIG: Optional[StorageConfig] = None


def _get_global_storage_config() -> StorageConfig:
    global _GLOBAL_STORAGE_CONFIG   # pylint: disable=global-statement
    if not _GLOBAL_STORAGE_CONFIG:
        _GLOBAL_STORAGE_CONFIG = get_storage_config()
    return _GLOBAL_STORAGE_CONFIG


def open(filename: str, mode: str = "rt"):  # pylint: disable=redefined-builtin
    """Open file using internal configuration"""
    cld_filename = f"{_get_global_storage_config().prefix}{filename}"
    return smart_open.open(cld_filename, mode)


def get_file_stream(file_path: str, preview: bool = False, file_type=None) -> BinaryIO:  # pylint: disable=unused-argument
    """Get file stream object (deprecated)"""
    # TODO: remove this function
    return open(file_path, "rb")


def upload_file_stream(fp: BinaryIO, file_path: Optional[str] = None, seek: bool = True) -> str:  # pylint: disable=invalid-name
    """Upload file stream to a given path (deprecated)"""
    # TODO: remove this function
    if seek:
        fp.seek(0)
    if file_path is None:
        file_path = str(uuid.uuid1())
    with open(file_path, "wb") as fo:   # pylint: disable=invalid-name
        fo.write(fp.read())
    return file_path



"""Cache Manager and utils."""
import datetime
import logging
from typing import Any, Dict, List, Optional, Union

from plynx.constants import NodeRunningStatus
from plynx.db.node import Node
from plynx.db.node_cache import NodeCache
from plynx.utils.common import ObjectId
from plynx.utils.db_connector import get_db_connector


class NodeCacheManager:
    """The Node cache interface.

    The cache is defined by Node's
        - original_node_id
        - inputs
        - parameters
    """

    @staticmethod
    def get(node: Node) -> Optional[NodeCache]:
        """Pull NodeCache if exists.

        Args:
            node        (Node):             Node object

        Return:
            (NodeCache)     NodeCache or None
        """
        key = NodeCache.generate_key(node)
        db_node_cache = get_db_connector().node_cache.find({
            'key': key,
            'removed': {'$ne': True}
        }).sort('insertion_date', -1).limit(1)
        caches = list(db_node_cache)
        if len(caches) > 0:
            return NodeCache.from_dict(caches[0])
        else:
            return None

    @staticmethod
    def post(node: Node, run_id: ObjectId) -> bool:
        """Create NodeCache instance in the database.

        Args:
            node        (Node):             Node object
            run_id      (ObjectId, str):    Run ID

        Return:
            True if cache saved else False
        """
        assert node.node_running_status == NodeRunningStatus.SUCCESS, \
            'Only Nodes with status SUCCESS can be cached'
        node_cache = NodeCache.instantiate(node=node, run_id=run_id)
        try:
            node_cache.save()
        except Exception as e:  # pylint: disable=broad-except
            logging.error(f"Could not save cache: `{e}`")
            return False
        return True

    @staticmethod
    def _make_query(
            start_datetime: Optional[datetime.datetime] = None,
            end_datetime: Optional[datetime.datetime] = None,
            non_protected_only: bool = False,
            ) -> Dict[str, Any]:
        """Make sample query.

        Args:
            start_datetime  (datetime, None):   Start datetime or None if selecting from beginning
            end_datetime    (datetime, None):   End datetime or None if selecting until now

        Return:
            Iterator on the list of dict-like objects
        """
        and_query: List[Dict[str, Dict[str, Union[bool, datetime.datetime]]]] = []

        insertion_query: Dict[str, Union[bool, datetime.datetime]] = {}
        if start_datetime:
            insertion_query['$gte'] = start_datetime
        if end_datetime:
            insertion_query['$lt'] = end_datetime
        if insertion_query:
            and_query.append({'insertion_date': insertion_query})

        if non_protected_only:
            and_query.append({'protected': {'$ne': True}})

        return {'$and': and_query} if and_query else {}

    @staticmethod
    def get_list(
            start_datetime: Optional[datetime.datetime] = None,
            end_datetime: Optional[datetime.datetime] = None,
            non_protected_only: bool = False
            ):
        """List of NodeCache objects.

        Args:
            start_datetime  (datetime, None):   Start datetime or None if selecting from beginning
            end_datetime    (datetime, None):   End datetime or None if selecting until now

        Return:
            Iterator on the list of dict-like objects
        """
        return get_db_connector().node_cache.find(NodeCacheManager._make_query(start_datetime, end_datetime, non_protected_only))

    @staticmethod
    def clean_up():
        """Remove NodeCache objects with flag `removed` set
        """
        return get_db_connector().node_cache.remove({'removed': True})



"""Node collection manager and utils"""

import logging
from collections import OrderedDict
from typing import Any, Dict, List, Optional, Union

from past.builtins import basestring
from pymongo import ReturnDocument

from plynx.constants import Collections, HubSearchParams, NodeRunningStatus, NodeStatus
from plynx.db.node import Node
from plynx.utils.common import ObjectId, parse_search_string, to_object_id
from plynx.utils.db_connector import get_db_connector
from plynx.utils.hub_node_registry import registry

_PROPERTIES_TO_GET_FROM_SUBS = ['node_running_status', 'logs', 'outputs', 'cache_url']


class NodeCollectionManager:
    """NodeCollectionManager contains all the operations to work with Nodes in the database."""

    def __init__(self, collection: str):
        super().__init__()

        self.collection: str = collection

    # pylint: disable=too-many-arguments
    def get_db_objects(
            self,
            status: Union[List[str], str] = '',
            node_kinds: Union[None, str, List[str]] = None,
            search: str = '',
            per_page: int = 20,
            offset: int = 0,
            user_id: Optional[ObjectId] = None,
            ) -> Optional[List[Dict]]:
        """Get subset of the Objects.

        Args:
            status              (str, None):                Node Running Status
            search              (str, None):                Search pattern
            per_page            (int):                      Number of Nodes per page
            offset              (int):                      Offset

        Return:
            (list of dict)  List of Nodes in dict format
        """
        # pylint: disable=too-many-branches
        if status and isinstance(status, basestring):
            status = [status]
        if node_kinds and isinstance(node_kinds, basestring):
            node_kinds = [node_kinds]

        aggregate_list: List[Dict[str, Any]] = []
        search_parameters, search_string = parse_search_string(search)

        # Match
        and_query: Dict[str, Union[ObjectId, Dict[str, Union[str, List[str], Dict]]]] = {}
        if node_kinds:
            and_query['kind'] = {'$in': node_kinds}
        if status:
            and_query['node_status'] = {'$in': status}
        if search_string:
            and_query['$text'] = {'$search': search_string}
        if 'original_node_id' in search_parameters:
            and_query['original_node_id'] = to_object_id(search_parameters['original_node_id'])
        if HubSearchParams.INPUT_FILE_TYPE in search_parameters:
            and_query['inputs'] = {"$elemMatch": {"file_type": search_parameters[HubSearchParams.INPUT_FILE_TYPE]}}
        if HubSearchParams.OUTPUT_FILE_TYPE in search_parameters:
            and_query['outputs'] = {"$elemMatch": {"file_type": search_parameters[HubSearchParams.OUTPUT_FILE_TYPE]}}
        if len(and_query) > 0:
            aggregate_list.append({"$match": and_query})

        # Join with users
        aggregate_list.append({
            '$lookup': {
                'from': 'users',
                'localField': 'author',
                'foreignField': '_id',
                'as': '_user'
            }
        })
        # rm password hash
        aggregate_list.append({
          "$project": {
            "_user.password_hash": 0,
          }
        })

        # Match username
        and_query = {}
        if 'author' in search_parameters:
            and_query['_user.username'] = search_parameters['author']

        if len(and_query):
            aggregate_list.append({"$match": and_query})

        # sort
        sort_dict = OrderedDict()
        if 'sort' in search_parameters:
            # TODO more sort options
            if search_parameters['sort'] == 'starred':
                sort_dict['starred'] = -1
        sort_dict['insertion_date'] = -1

        aggregate_list.append({
            "$sort": sort_dict
            }
        )
        aggregate_list.append({
            "$addFields": {
                '_readonly': {'$ne': ["$author", to_object_id(user_id)]},
            }
        })
        # counts and pagination
        aggregate_list.append({
            '$facet': {
                "metadata": [{"$count": "total"}],
                "list": [{"$skip": int(offset)}, {"$limit": int(per_page)}],
            }
        })

        # TODO: remove `allowDiskUse=True` and update indexes instead
        return next(get_db_connector()[self.collection].aggregate(aggregate_list, allowDiskUse=True), None)

    def get_db_objects_by_ids(self, ids: Union[List[ObjectId], List[str]], collection: Optional[str] = None) -> List[dict]:
        """Find all the Objects with a given IDs.

        Args:
            ids    (list of ObjectID):  Object Ids
        """
        if collection == Collections.HUB_NODE_REGISTRY:
            # TODO separate types of ids into different functions
            db_objects = map(lambda node: node.to_dict(), registry.find_nodes(ids))     # type: ignore
        else:
            db_objects = get_db_connector()[collection or self.collection].find({
                '_id': {
                    '$in': list(ids)
                }
            })

        return list(db_objects)

    def _update_sub_nodes_fields(
            self,
            sub_nodes_dicts: List[Dict],
            reference_node_id: str,
            target_props: List[str],
            reference_collection: Optional[str] = None
            ):
        reference_collection = reference_collection or self.collection
        id_to_updated_node_dict = {}
        function_location_to_updated_node_dict = {}
        upd_node_ids = list(map(lambda node_dict: node_dict.get(reference_node_id, "unknown"), sub_nodes_dicts))
        for upd_node_dict in self.get_db_objects_by_ids(upd_node_ids, collection=reference_collection):
            id_to_updated_node_dict[upd_node_dict['_id']] = upd_node_dict
            function_location_to_updated_node_dict[upd_node_dict.get("code_function_location", "unknown")] = upd_node_dict
        for sub_node_dict in sub_nodes_dicts:
            if sub_node_dict.get(reference_node_id, "unknown") not in id_to_updated_node_dict:
                continue
            for prop in target_props:
                sub_node_dict[prop] = id_to_updated_node_dict[sub_node_dict.get(reference_node_id, "unknown")][prop]

        if reference_collection == Collections.HUB_NODE_REGISTRY:
            # special case: we need to compare not target_props, but rather assign it
            assert len(target_props) == 1, "Only node_status can be assigned"
            assert target_props[0] == 'node_status', "Only node_status can be assigned"
            for sub_node_dict in sub_nodes_dicts:
                if sub_node_dict.get(reference_node_id, "unknown") is None:
                    continue
                if sub_node_dict.get(reference_node_id, "unknown") not in function_location_to_updated_node_dict:
                    logging.warning(f"`{sub_node_dict.get(reference_node_id, 'unknown')}` is not found in the list of operation locations")
                    continue
                if sub_node_dict['code_hash'] != function_location_to_updated_node_dict[sub_node_dict.get(reference_node_id, "unknown")]["code_hash"]:
                    sub_node_dict['node_status'] = NodeStatus.DEPRECATED

    def get_db_node(self, node_id: ObjectId, user_id: Optional[ObjectId] = None):
        """Get dict representation of a Node.

        Args:
            node_id     (ObjectId, str):        Object ID
            user_id     (str, ObjectId, None):  User ID

        Return:
            (dict)  dict representation of the Object
        """
        res = self.get_db_object(node_id, user_id)
        if not res:
            return res

        sub_nodes_dicts: Optional[List[Dict]] = None
        for parameter in res['parameters']:
            if parameter['name'] == '_nodes':
                sub_nodes_dicts = parameter['value']['value']
                break

        if sub_nodes_dicts:
            # TODO join collections using database capabilities
            if self.collection == Collections.RUNS:
                self._update_sub_nodes_fields(sub_nodes_dicts, '_id', _PROPERTIES_TO_GET_FROM_SUBS)
            self._update_sub_nodes_fields(sub_nodes_dicts, 'original_node_id', ['node_status'], reference_collection=Collections.TEMPLATES)
            self._update_sub_nodes_fields(sub_nodes_dicts, 'code_function_location', ['node_status'], reference_collection=Collections.HUB_NODE_REGISTRY)

        return res

    def get_db_object(self, object_id: ObjectId, user_id: Optional[ObjectId] = None) -> Dict:
        """Get dict representation of an Object.

        Args:
            object_id   (ObjectId):        Object ID
            user_id     (ObjectId, None):  User ID

        Return:
            (dict)  dict representation of the Object
        """
        res = get_db_connector()[self.collection].find_one({'_id': to_object_id(object_id)})
        if not res:
            return res

        res['_readonly'] = (user_id != to_object_id(res['author']))

        return res

    @staticmethod
    def _transplant_node(node: Node, dest_node: Node) -> Node:
        if dest_node._id == node.original_node_id:
            return node
        dest_node.apply_properties(node)
        dest_node.original_node_id = dest_node._id
        dest_node.parent_node_id = dest_node.successor_node_id = None
        dest_node._id = node._id
        return dest_node

    def upgrade_sub_nodes(self, main_node: Node) -> int:
        """Upgrade deprecated Nodes.

        The function does not change the original graph in the database.

        Return:
            (int):  Number of upgraded Nodes
        """
        assert self.collection == Collections.TEMPLATES
        sub_nodes = main_node.get_parameter_by_name('_nodes').value.value
        node_ids = [node.original_node_id for node in sub_nodes]
        db_nodes = self.get_db_objects_by_ids(node_ids)
        new_node_db_mapping = {}

        for db_node in db_nodes:
            original_node_id = db_node['_id']
            new_db_node = db_node
            if original_node_id not in new_node_db_mapping:
                while new_db_node['node_status'] != NodeStatus.READY and 'successor_node_id' in new_db_node and new_db_node['successor_node_id']:
                    tmp_node = self.get_db_node(new_db_node['successor_node_id'])
                    if tmp_node:
                        new_db_node = tmp_node
                    else:
                        break
                new_node_db_mapping[original_node_id] = new_db_node

        new_nodes = [
            NodeCollectionManager._transplant_node(
                node,
                Node.from_dict(new_node_db_mapping[to_object_id(node.original_node_id)])
            ) for node in sub_nodes]

        upgraded_nodes_count = sum(
            1 for node, new_node in zip(sub_nodes, new_nodes) if node.original_node_id != new_node.original_node_id
        )

        main_node.get_parameter_by_name('_nodes').value.value = new_nodes
        return upgraded_nodes_count

    def pick_node(self, kinds: List[str]) -> Dict:
        """Get node and set status to RUNNING in atomic way"""
        node = get_db_connector()[self.collection].find_one_and_update(
            {
                '$and': [
                    {
                        'kind': {
                            '$in': kinds,
                        }
                    },
                    {
                        'node_running_status': {
                            '$in': [
                                NodeRunningStatus.READY,
                                NodeRunningStatus.IN_QUEUE,
                            ]
                        }
                    },
                ],
            },
            {
                '$set': {
                    'node_running_status': NodeRunningStatus.RUNNING
                }
            },
            return_document=ReturnDocument.AFTER
        )
        return node



"""Main PLynx constants defined in this module"""
# flake8: noqa
from plynx.constants.collections import Collections
from plynx.constants.node_enums import (IGNORED_CACHE_PARAMETERS, NodeClonePolicy, NodeOrigin, NodePostAction, NodePostStatus, NodeRunningStatus, NodeStatus,
                                        NodeVirtualCollection, SpecialNodeId)
from plynx.constants.parameter_types import ParameterTypes
from plynx.constants.resource_enums import HubSearchParams, NodeResources
from plynx.constants.users import IAMPolicies, RegisterUserExceptionCode, TokenType, UserPostAction
from plynx.constants.validation_enums import ValidationCode, ValidationTargetType
from plynx.constants.web import ResponseStatus



"""Plynx standard Hub based on the fixed list of Operations"""

import copy
import pydoc
from typing import Any, Callable, Dict, List, Union

import plynx.node
from plynx.base import hub
from plynx.constants import HubSearchParams
from plynx.db.node import Node
from plynx.utils.common import parse_search_string
from plynx.utils.hub_node_registry import registry


def _enhance_list_item(raw_item: Dict) -> Dict:
    if raw_item['_type'] == 'Group':
        # TODO proper checking
        items = []
        for raw_subitem in raw_item['items']:
            items.append(_enhance_list_item(raw_subitem))
        raw_item['items'] = items
        return raw_item
    # check if the node is valid
    node = Node.from_dict(raw_item)
    registry.register_node(node)
    return node.to_dict()


def _recursive_filter(search_parameters: Dict[str, str], search_string: str, list_of_nodes: List[Dict]):
    res = []
    for raw_node in list_of_nodes:
        if raw_node["_type"] == "Group":
            raw_group = copy.deepcopy(raw_node)
            raw_group['items'] = _recursive_filter(search_parameters, search_string, raw_group['items'])
            res.append(raw_group)
        elif len(search_string) == 0 or search_string.upper() in raw_node['title'].upper():
            input_file_type = search_parameters.get(HubSearchParams.INPUT_FILE_TYPE, None)
            output_file_type = search_parameters.get(HubSearchParams.OUTPUT_FILE_TYPE, None)
            if input_file_type:
                for input in raw_node["inputs"]:    # pylint: disable=redefined-builtin
                    if input["file_type"] == input_file_type:
                        break
                else:
                    continue
            if output_file_type:
                for output in raw_node["outputs"]:
                    if output["file_type"] == output_file_type:
                        break
                else:
                    continue
            res.append(raw_node)
    return res


class StaticListHub(hub.BaseHub):
    """Plynx standard Hub based on the fixed list of Operations"""
    def __init__(self, list_module: str):
        super().__init__()

        collection: List[Union[Callable, plynx.node.utils.Group]] = pydoc.locate(list_module)   # type: ignore

        assert collection is not None, f"Module `{list_module}` not found"

        self.list_of_nodes = []
        for func_or_group in collection:
            raw_item = plynx.node.utils.func_or_group_to_dict(func_or_group)
            self.list_of_nodes.append(_enhance_list_item(raw_item))

    def search(self, query: hub.Query) -> Dict[str, Any]:
        # TODO use search_parameters
        # TODO should parse_search_string be removed from nodes_collection?
        search_parameters, search_string = parse_search_string(query.search)

        res = _recursive_filter(search_parameters, search_string, self.list_of_nodes)
        return {
            'list': res,
            'metadata': [
                {
                    'total': len(res)
                },
            ],
        }



"""Resource enums"""


class NodeResources:
    """Internal node elements"""
    INPUT: str = 'inputs'
    OUTPUT: str = 'outputs'
    CLOUD_INPUT: str = 'cloud_inputs'
    CLOUD_OUTPUT: str = 'cloud_outputs'
    PARAM: str = 'params'
    LOG: str = 'logs'


class HubSearchParams:
    """Describing serach based on resource"""
    INPUT_FILE_TYPE: str = "input_file_type"
    OUTPUT_FILE_TYPE: str = "output_file_type"



"""Standard Executors that support running on local machine."""
import logging
import os
import signal
import threading
from abc import abstractmethod
from collections import defaultdict
from subprocess import Popen
from typing import Any, Dict, List, Optional, Union

import jinja2
from past.builtins import basestring

import plynx.plugins.executors.bases
import plynx.utils.plugin_manager
from plynx.constants import NodeResources, NodeRunningStatus, ParameterTypes
from plynx.db.node import Node, Output, Parameter, ParameterCode
from plynx.plugins.resources.common import FILE_KIND
from plynx.utils.file_handler import get_file_stream, upload_file_stream


def _resource_merger_func():
    return defaultdict(list)


def prepare_parameters_for_python(parameters: List[Parameter]) -> Dict[str, Any]:
    """Pythonize parameters"""
    res = {}
    for parameter in parameters:
        value = None
        if parameter.parameter_type == ParameterTypes.ENUM:
            index = max(0, min(len(parameter.value.values) - 1, int(parameter.value.index)))
            value = parameter.value.values[index]
        elif parameter.parameter_type in [ParameterTypes.LIST_STR, ParameterTypes.LIST_INT]:
            if parameter.parameter_type == ParameterTypes.LIST_INT:
                value = list(map(int, parameter.value))
            else:
                value = parameter.value
        elif parameter.parameter_type == ParameterTypes.CODE:
            value = parameter.value.value
        elif parameter.parameter_type == ParameterTypes.INT:
            value = int(parameter.value)
        elif parameter.parameter_type == ParameterTypes.FLOAT:
            value = float(parameter.value)
        else:
            value = parameter.value
        res[parameter.name] = value
    return res


class _ResourceMerger:
    # TODO rename arguments
    def __init__(self, init_level_0: List[str], init_level_1: List[str]):
        self._dict: Dict[str, Dict[str, Union[List[str], str]]] = defaultdict(_resource_merger_func)
        for key in init_level_0:
            self._dict[key] = _resource_merger_func()
            for lev_1 in init_level_1:
                self._dict[key][lev_1] = []

    def append(self, resource_dict: Dict[str, List[str]], resource_name: str, is_list: bool):
        """Append values to the resource"""
        for key, value in resource_dict.items():
            if is_list:
                self._dict[key][resource_name].append(value)    # type: ignore
            else:
                self._dict[key][resource_name] = value

    def get_dict(self) -> Dict[str, Dict[str, Union[List[str], str]]]:
        """
        Return original dict.

        .. highlight:: python
        .. code-block:: python
        .. {
        ..     'inputs': {
        ..         'input_name_0': [],
        ..         'input_name_1': ['/tmp/1', '/tmp/2'],
        ..         'input_name_2': '/tmp/3',
        ..     },
        ..     'cloud_inputs': {
        ..         'input_name_0': [],
        ..         'input_name_1': ['gs://1', 'gs://2'],
        ..         'input_name_2': 'gs://3',
        ..     }
        .. }
        ..

        Out:    Dict
        """
        return self._dict


class BaseBash(plynx.plugins.executors.bases.PLynxAsyncExecutorWithDirectory):
    """Base Executor that will use unix bash as a backend."""
    # pylint: disable=too-many-instance-attributes

    def __init__(self, node: Optional[Node] = None):
        super().__init__(node)
        self.sp: Optional[Popen] = None
        self.logs_sizes: Dict[str, int] = {}
        self.final_logs_uploaded = False
        self.logs: Dict[str, str] = {}
        self.logs_lock = threading.Lock()
        self.output_to_filename: Dict[str, str] = {}
        self._resource_manager = plynx.utils.plugin_manager.get_resource_manager()
        self._command = 'bash'
        self._node_running_status = NodeRunningStatus.READY

    def exec_script(self, script_location: str) -> str:
        """Execute the script when inputs are initialized."""
        self._node_running_status = NodeRunningStatus.SUCCESS

        try:
            def pre_exec():
                # Restore default signal disposition and invoke setsid
                for sig in ('SIGPIPE', 'SIGXFZ', 'SIGXFSZ'):
                    if hasattr(signal, sig):
                        signal.signal(getattr(signal, sig), signal.SIG_DFL)
                os.setsid()

            env = os.environ.copy()

            # append running script to worker log
            with open(script_location, 'r') as script_f, open(self.logs['worker'], 'a') as worker_f:
                worker_f.write(self._make_debug_text("Running script:"))
                worker_f.write(script_f.read())
                worker_f.write('\n')
                worker_f.write(self._make_debug_text("End script"))

            with open(self.logs['stdout'], 'wb') as stdout_file, open(self.logs['stderr'], 'wb') as stderr_file:
                self.sp = Popen(    # pylint: disable=subprocess-popen-preexec-fn,consider-using-with
                    [self._command, script_location],
                    stdout=stdout_file, stderr=stderr_file,
                    cwd=self.workdir, env=env,
                    preexec_fn=pre_exec)

                assert self.sp, "Popen object was not initialized"
                self.sp.wait()

            if self.sp.returncode:
                raise Exception("Process returned non-zero value")

        except Exception as e:  # pylint: disable=broad-except
            if self._node_running_status != NodeRunningStatus.CANCELED:
                self._node_running_status = NodeRunningStatus.FAILED
            logging.exception("Job failed")
            with open(self.logs['worker'], 'a+') as worker_log_file:
                worker_log_file.write(self._make_debug_text("JOB FAILED"))
                worker_log_file.write(str(e))

        return self._node_running_status

    def kill(self):
        if not hasattr(self, 'sp') or not self.sp:
            return

        self._node_running_status = NodeRunningStatus.CANCELED

        logging.info('Sending SIGTERM signal to bash process group')
        try:
            os.killpg(os.getpgid(self.sp.pid), signal.SIGTERM)
            logging.info(f"Killed %d{self.sp.pid}")
        except OSError as e:
            logging.error(f"Error: {e}")

    def is_updated(self):
        logging.info('Tick')
        return self.upload_logs(final=False)

    # Hack: do not pickle file
    def __getstate__(self):
        logging.critical('Run into `__getstate__`! Look ad `logs_lock` variable')
        dict_copy = dict(self.__dict__)
        if 'sp' in dict_copy:
            del dict_copy['sp']
        return dict_copy

    @staticmethod
    def _make_debug_text(text: str) -> str:
        content = "\n".join([f"# {line}" for line in text.split("\n")])
        border = "#" * 40
        return f"{border}\n{content}\n{border}\n"

    @classmethod
    def get_default_node(cls, is_workflow: bool) -> Node:
        if is_workflow:
            raise Exception('This class cannot be a workflow')
        node = super().get_default_node(is_workflow)
        node.parameters.extend(
            [
                Parameter(
                    name='_cmd',
                    parameter_type=ParameterTypes.CODE,
                    value=ParameterCode(
                        mode='sh',
                        value='set -e\n\necho "hello world"\n',
                    ),
                    mutable_type=False,
                    publicable=False,
                    removable=False,
                ),
                Parameter(
                    name='_cacheable',
                    parameter_type=ParameterTypes.BOOL,
                    value=False,
                    mutable_type=False,
                    publicable=False,
                    removable=False,
                ),
                Parameter(
                    name='_timeout',
                    parameter_type=ParameterTypes.INT,
                    value=600,
                    mutable_type=False,
                    publicable=True,
                    removable=False
                ),
            ]
        )
        node.logs.extend(
            [
                Output(
                    name='stderr',
                    file_type=FILE_KIND,
                ),
                Output(
                    name='stdout',
                    file_type=FILE_KIND,
                ),
                Output(
                    name='worker',
                    file_type=FILE_KIND,
                ),
            ]
        )
        return node

    def _prepare_inputs(self, preview: bool = False):
        assert self.node, "Attribute `node` is undefined"
        resource_merger = _ResourceMerger(
            [NodeResources.INPUT],
            [input.name for input in self.node.inputs if input.is_array],
        )
        for input in self.node.inputs:  # pylint: disable=redefined-builtin
            if preview:
                for i, _ in enumerate(range(input.min_count)):
                    filename = os.path.join(self.workdir, f"i_{i}_{input.name}")
                    resource_merger.append(
                        self._resource_manager.kind_to_resource_class[input.file_type].prepare_input(filename, preview),
                        input.name,
                        input.is_array,
                    )
            else:
                for i, value in enumerate(input.values):
                    filename = os.path.join(self.workdir, f"i_{i}_{input.name}")
                    with open(filename, 'wb') as f:
                        f.write(get_file_stream(value).read())
                    resource_merger.append(
                        self._resource_manager.kind_to_resource_class[input.file_type].prepare_input(filename, preview),
                        input.name,
                        input.is_array,
                    )
        return resource_merger.get_dict()

    def _prepare_outputs(self, preview: bool = False):
        assert self.node, "Attribute `node` is undefined"
        resource_merger = _ResourceMerger(
            [NodeResources.OUTPUT],
            [output.name for output in self.node.outputs if output.is_array],
        )
        for output in self.node.outputs:
            filename = os.path.join(self.workdir, f"o_{output.name}")
            self.output_to_filename[output.name] = filename
            resource_merger.append(
                self._resource_manager.kind_to_resource_class[output.file_type].prepare_output(filename, preview),
                output.name,
                is_list=False
            )
        return resource_merger.get_dict()

    def _prepare_logs(self):
        with self.logs_lock:
            self.logs = {}
            for log in self.node.logs:
                filename = os.path.join(self.workdir, f"l_{log.name}")
                self.logs[log.name] = filename
                self.logs_sizes[log.name] = 0
            return self.logs

    def _get_script_fname(self, extension: str = ".sh"):
        return os.path.join(self.workdir, f"exec{extension}")

    def _prepare_parameters(self):
        return prepare_parameters_for_python(self.node.parameters)

    def _postprocess_outputs(self, outputs: Dict[str, str]):
        assert self.node, "Attribute `node` is undefined"
        for key, filename in outputs.items():
            logging.info(f"Uploading output `{key}` - `{filename}`")
            if os.path.exists(filename):
                logging.info('path exists')
                matching_outputs = list(filter(lambda o: o.name == key, self.node.outputs))     # pylint: disable=cell-var-from-loop
                assert len(matching_outputs) == 1, f"Found more that 1 output with the same name `{key}`"
                filename = self._resource_manager.kind_to_resource_class[matching_outputs[0].file_type].postprocess_output(filename)
                logging.info(filename)
                with open(filename, 'rb') as f:
                    # resource_id
                    self.node.get_output_by_name(key).values = [upload_file_stream(f)]
                    logging.info(self.node.get_output_by_name(key).to_dict())
            else:
                raise IOError(f"Output `{key}` (filename: `{filename}`) does not exist")

    def _postprocess_logs(self) -> None:
        self.upload_logs(final=True)

    def _extract_cmd_text(self) -> str:
        assert self.node, "Attribute `node` is undefined"
        parameter = self.node.get_parameter_by_name('_cmd')
        if parameter.parameter_type == ParameterTypes.CODE:
            return parameter.value.value
        elif parameter.parameter_type == ParameterTypes.TEXT:
            # backward compatibility: `cmd` used to have type of TEXT
            return parameter.value
        raise TypeError("Process returned non-zero value")

    def upload_logs(self, final: bool = False) -> bool:
        """Upload logs to the storage. When Final is False, only upload on update"""
        assert self.node, "Attribute `node` is undefined"
        is_dirty = False
        with self.logs_lock:
            if self.final_logs_uploaded:
                return is_dirty
            self.final_logs_uploaded = final
            for key, filename in self.logs.items():
                if key not in self.logs_sizes:
                    # the logs have not been initialized yet
                    continue
                if os.path.exists(filename) and os.stat(filename).st_size != self.logs_sizes[key]:
                    is_dirty = True
                    log = self.node.get_log_by_name(key)
                    self.logs_sizes[key] = os.stat(filename).st_size
                    with open(filename, 'rb') as f:
                        # resource_id should be None if the file has not been uploaded yet
                        # otherwise assign it
                        log.values = [upload_file_stream(f, log.values[0] if len(log.values) > 0 else None)]
        return is_dirty

    @abstractmethod
    def run(self, preview=False):
        pass


class BashJinja2(BaseBash):
    """Local executor that uses jinja2 template to format a bash script."""
    HELP_TEMPLATE = """# Use templates: {}
# For example `{{{{ '{{{{' }}}} params['_timeout'] {{{{ '}}}}' }}}}` or `{{{{ '{{{{' }}}} inputs['abc'] {{{{ '}}}}' }}}}`

"""

    def run(self, preview: bool = False) -> str:
        inputs = self._prepare_inputs(preview)
        parameters = self._prepare_parameters()
        outputs = self._prepare_outputs(preview)
        logs = self._prepare_logs()
        if preview:
            help_msg = BashJinja2.HELP_TEMPLATE.format(list(inputs.keys()) + list(outputs.keys()) + [NodeResources.PARAM])
        else:
            help_msg = ""
        cmd = f"{help_msg}{self._extract_cmd_text()}"
        cmd_template = jinja2.Template(cmd)
        resources = inputs
        resources.update(outputs)
        cmd_string = cmd_template.render(
            params=parameters,
            logs=logs,
            **resources
        )
        if preview:
            return cmd_string

        script_location = self._get_script_fname()
        with open(script_location, 'w') as script_file:
            script_file.write(
                cmd_string
            )

        self._node_running_status = self.exec_script(script_location)

        self._postprocess_outputs(outputs[NodeResources.OUTPUT])
        self._postprocess_logs()

        return self._node_running_status

    @classmethod
    def get_default_node(cls, is_workflow: bool) -> Node:
        node = super().get_default_node(is_workflow)
        node.title = 'New bash script'
        return node


class PythonNode(BaseBash):
    """Local executor that uses python template to format a bash script."""
    def __init__(self, node: Optional[Node] = None):
        super().__init__(node)
        self._command = 'python'

    def run(self, preview: bool = False) -> str:
        inputs = self._prepare_inputs(preview)
        parameters = self._prepare_parameters()
        outputs = self._prepare_outputs(preview)
        logs = self._prepare_logs()
        cmd = self._extract_cmd_text()
        cmd_array = []
        cmd_array.extend([
            self._get_arguments_string(key, value)
            for key, value in inputs.items()
        ])
        cmd_array.extend([
            self._get_arguments_string(key, value)
            for key, value in outputs.items()
        ])
        cmd_array.extend([
            self._get_arguments_string(NodeResources.PARAM, parameters),
            self._get_arguments_string(NodeResources.LOG, logs),
            "\n",
            "# User code starts there:",
            cmd,
        ])
        cmd_string = '\n'.join(cmd_array)

        if preview:
            return cmd_string

        script_location = self._get_script_fname(extension='.py')
        with open(script_location, 'w') as script_file:
            script_file.write(
                cmd_string
            )

        res = self.exec_script(script_location)

        self._postprocess_outputs(outputs[NodeResources.OUTPUT])
        self._postprocess_logs()

        return res

    @classmethod
    def _get_arguments_string(cls, var_name: str, arguments: Dict[str, Any]) -> str:
        res = [f"{var_name} = {{}}"]
        for key, value in arguments.items():
            value = repr(cls._pythonize(value))
            res.append(f'{var_name}["{key}"] = {value}')
        return '\n'.join(res)

    @staticmethod
    def _pythonize(value: Any) -> Any:
        if isinstance(value, basestring):
            repr(value)
        return value

    @classmethod
    def get_default_node(cls, is_workflow: bool) -> Node:
        node = super().get_default_node(is_workflow)
        node.title = 'New python script'
        param = list(filter(lambda p: p.name == '_cmd', node.parameters))[0]
        param.value.mode = 'python'
        param.value.value = 'print("hello world")'
        return node


class File(plynx.plugins.executors.bases.PLynxAsyncExecutor):
    """Dummy executor that represents STATIC Operations."""

    def run(self, preview: bool = False):
        if preview:
            return "Cannot preview the content. Please check the outputs."
        raise NotImplementedError()

    def kill(self):
        raise NotImplementedError()

    @classmethod
    def get_default_node(cls, is_workflow: bool) -> Node:
        raise NotImplementedError()



"""Python Operation"""
import contextlib
import inspect
import os
import sys
import threading
import uuid
from typing import Any, Callable, Dict

import cloudpickle

import plynx.plugins.executors.bases
import plynx.plugins.executors.local
import plynx.utils.plugin_manager
from plynx.constants import NodeRunningStatus, ParameterTypes
from plynx.db.node import Input, Node, Output, Parameter, ParameterCode

DEFAULT_CMD = """# Python Operation
# The code of the operation have to be defined as a function
#
# A function must be named `operation`.
# The arguments are named and must represent then inputs of the PLynx Operation.
# The function must return a dictionary that map to the outputs of PLynx Operation
def operation(int_a, int_b):
    return {"sum": int_a + int_b}
"""

stateful_init_mutex = threading.Lock()
stateful_class_registry = {}

_resource_manager = plynx.utils.plugin_manager.get_resource_manager()


def materialize_fn_or_cls(node: Node) -> Callable:
    """Unpickle the function"""

    pickled_fn_parameter = node.get_parameter_by_name("_pickled_fn", throw=False)
    code_parameter = node.get_parameter_by_name("_cmd", throw=False)
    assert not (pickled_fn_parameter and code_parameter), "`_pickled_fn` and `_cmd` cannot be both non-null"
    if pickled_fn_parameter:
        fn_str: str = pickled_fn_parameter.value

        func_bytes = bytes.fromhex(fn_str)
        func = cloudpickle.loads(func_bytes)
        return func
    elif code_parameter:
        code = code_parameter.value.value
        local_vars: Dict[str, Any] = {}
        exec(code, globals(), local_vars)   # pylint: disable=W0122
        return local_vars["operation"]
    raise ValueError("No function to materialize")


def assign_outputs(node: Node, output_dict: Dict[str, Any]):
    """Apply output_dict to node's outputs."""
    if not output_dict:
        return
    for key, value in output_dict.items():
        node_output = node.get_output_by_name(key)
        func = _resource_manager.kind_to_resource_class[node_output.file_type].postprocess_output
        if node_output.is_array:
            node_output.values = list(map(func, value))
        else:
            node_output.values = [func(value)]


class redirect_to_plynx_logs:   # pylint: disable=invalid-name
    """Redirect stdout and stderr to standard PLynx Outputs"""
    # pylint: disable=too-many-instance-attributes

    def __init__(self, node: Node, stdout: str, stderr: str):
        self.stdout_filename = str(uuid.uuid4())
        self.stderr_filename = str(uuid.uuid4())

        self.stdout = stdout
        self.stderr = stderr

        self.node = node

        self.names_map = [
            (self.stdout_filename, self.stdout),
            (self.stderr_filename, self.stderr),
        ]

        self.stdout_file = None
        self.stderr_file = None

        self.stdout_redirect = None
        self.stderr_redirect = None

    def __enter__(self):
        self.stdout_file = open(self.stdout_filename, 'w')
        self.stderr_file = open(self.stderr_filename, 'w')

        self.stdout_redirect = contextlib.redirect_stdout(self.stdout_file)
        self.stderr_redirect = contextlib.redirect_stderr(self.stderr_file)

        self.stdout_redirect.__enter__()
        self.stderr_redirect.__enter__()

    def __exit__(self, *args):

        self.stdout_redirect.__exit__(*sys.exc_info())
        self.stderr_redirect.__exit__(*sys.exc_info())

        self.stdout_file.close()
        self.stderr_file.close()

        for filename, logs_name in self.names_map:
            if os.stat(filename).st_size > 0:
                with plynx.utils.file_handler.open(filename, "w") as f:
                    with open(filename) as fi:
                        f.write(fi.read())
                output = self.node.get_log_by_name(name=logs_name)
                output.values = [filename]


def prep_args(node: Node) -> Dict[str, Any]:
    """Pythonize inputs and parameters"""
    args = {}
    for input in node.inputs:   # pylint: disable=redefined-builtin
        func = _resource_manager.kind_to_resource_class[input.file_type].preprocess_input
        if input.is_array:
            args[input.name] = list(map(func, input.values))
        else:
            args[input.name] = func(input.values[0])

    # TODO smater way to determine what parameters to pass
    visible_parameters = list(filter(lambda param: param.widget is not None, node.parameters))
    args.update(
        plynx.plugins.executors.local.prepare_parameters_for_python(visible_parameters)
    )
    return args


class PythonNode(plynx.plugins.executors.bases.PLynxSyncExecutor):
    """
    Class is used as a placeholder for local python executor
    """

    def run(self, preview: bool = False) -> str:
        assert self.node, "Executor memeber `node` is not defined"
        func = materialize_fn_or_cls(self.node)
        if inspect.isclass(func):
            with stateful_init_mutex:
                if self.node.code_hash not in stateful_class_registry:
                    with redirect_to_plynx_logs(self.node, "init_stdout", "init_stderr"):
                        stateful_class_registry[self.node.code_hash] = func()
                func = stateful_class_registry[self.node.code_hash]

        with redirect_to_plynx_logs(self.node, "stdout", "stderr"):
            res = func(**prep_args(self.node))

        assign_outputs(self.node, res)

        return NodeRunningStatus.SUCCESS

    def kill(self):
        raise NotImplementedError()

    @classmethod
    def get_default_node(cls, is_workflow: bool) -> Node:
        """Generate a new default Node for this executor"""
        if is_workflow:
            raise Exception('This class cannot be a workflow')
        node = super().get_default_node(is_workflow)
        node.inputs.extend(
            [
                Input(
                    name="int_a",
                    file_type="int",
                ),
                Input(
                    name="int_b",
                    file_type="int",
                ),
            ]
        )
        node.outputs.extend(
            [
                Output(
                    name="sum",
                    file_type="int",
                ),
            ]
        )
        node.parameters.extend(
            [
                Parameter(
                    name="_cmd",
                    parameter_type=ParameterTypes.CODE,
                    value=ParameterCode(
                        mode="python",
                        value=DEFAULT_CMD,
                    ),
                    mutable_type=False,
                    publicable=False,
                    removable=False,
                ),
            ]
        )
        return node



"""Templates for PLynx Resources and utils."""
from collections import namedtuple
from typing import Any, Dict, Optional

from plynx.constants import NodeResources

PreviewObject = namedtuple('PreviewObject', ['fp', 'resource_id'])


def _force_decode(byte_array):
    try:
        return byte_array.decode("utf-8")
    except UnicodeDecodeError:
        return f"# not a UTF-8 sequence:\n{byte_array}"
    return "Failed to decode the sequence"


class BaseResource:
    """Base Resource class"""
    DISPLAY_RAW: bool = False
    DISPLAY_THUMBNAIL: bool = False

    def __init__(self):
        pass

    @staticmethod
    def prepare_input(filename: str, preview: bool = False) -> Dict[str, str]:     # pylint: disable=unused-argument
        """Resource preprocessor"""
        return {NodeResources.INPUT: filename}

    @staticmethod
    def prepare_output(filename: str, preview: bool = False) -> Dict[str, str]:
        """Prepare output"""
        if not preview:
            # Create file
            with open(filename, 'a'):
                pass
        return {NodeResources.OUTPUT: filename}

    @staticmethod
    def preprocess_input(value: Any) -> Any:
        """Resource preprocessor"""
        return value

    @staticmethod
    def postprocess_output(value: Any) -> Any:
        """Resource postprocessor"""
        return value

    @classmethod
    def preview(cls, preview_object: PreviewObject) -> str:
        """Preview Resource"""
        # TODO escape html code for security reasons
        data = _force_decode(preview_object.fp.read())
        return f"<pre>{data}</pre>"

    @classmethod
    def thumbnail(cls, output: Any) -> Optional[str]:   # pylint: disable=unused-argument
        """Thumbnail preview Resource"""
        return None



"""Thumbnail utils"""
import logging
from typing import Optional

import plynx.db.node
import plynx.utils.plugin_manager
from plynx.constants import NodeRunningStatus


def get_thumbnail(output: plynx.db.node.Output) -> Optional[str]:
    """Apply a single thumbnail"""
    cls = plynx.utils.plugin_manager.get_resource_manager().kind_to_resource_class[output.file_type]
    if not cls.DISPLAY_THUMBNAIL:
        return None
    return cls.thumbnail(output)


def apply_thumbnails(node: plynx.db.node.Node):
    """
    Fill thumbnail field of every subnode
    """
    sub_nodes_parameter = node.get_parameter_by_name('_nodes', throw=False)
    if not sub_nodes_parameter:
        logging.warning("no subnodes found in `apply_thumbnails`")
        return

    sub_nodes = sub_nodes_parameter.value.value

    for sub_node in sub_nodes:
        if NodeRunningStatus.is_succeeded(sub_node.node_running_status):
            for output in sub_node.outputs:
                output.thumbnail = get_thumbnail(output)
        if sub_node._cached_node:
            for output in sub_node._cached_node.outputs:
                output.thumbnail = get_thumbnail(output)



"""Utils to work with executors"""
from typing import Any, Dict, Union

import plynx.base.executor
import plynx.db.node
import plynx.utils.exceptions
import plynx.utils.plugin_manager


def materialize_executor(node: Union[Dict[str, Any], plynx.db.node.Node]) -> plynx.base.executor.BaseExecutor:
    """
    Create a Node object from a dictionary

    Parameters:
        node: dictionary representation of a Node or the node itself

    Returns:
        Executor: Executor based on the kind of the Node and the config
    """

    if isinstance(node, dict):
        node = plynx.db.node.Node.from_dict(node)
    cls = plynx.utils.plugin_manager.get_executor_manager().kind_to_executor_class.get(node.kind)
    if not cls:
        raise plynx.utils.exceptions.ExecutorNotFound(
            f"Node kind `{node.kind}` not found"
        )
    return cls(node)



"""User based enums"""


class IAMPolicies:
    """Standard role policies"""
    CAN_VIEW_OTHERS_OPERATIONS: str = 'CAN_VIEW_OTHERS_OPERATIONS'
    CAN_VIEW_OTHERS_WORKFLOWS: str = 'CAN_VIEW_OTHERS_WORKFLOWS'
    CAN_VIEW_OPERATIONS: str = 'CAN_VIEW_OPERATIONS'
    CAN_VIEW_WORKFLOWS: str = 'CAN_VIEW_WORKFLOWS'

    CAN_CREATE_OPERATIONS: str = 'CAN_CREATE_OPERATIONS'
    CAN_CREATE_WORKFLOWS: str = 'CAN_CREATE_WORKFLOWS'

    CAN_MODIFY_OTHERS_WORKFLOWS: str = 'CAN_MODIFY_OTHERS_WORKFLOWS'

    CAN_RUN_WORKFLOWS: str = 'CAN_RUN_WORKFLOWS'

    IS_ADMIN: str = 'IS_ADMIN'


class UserPostAction:
    """HTTP POST action options"""
    MODIFY: str = 'MODIFY'


class RegisterUserExceptionCode:
    """Validation error codes"""
    EMPTY_USERNAME: str = 'EMPTY_USERNAME'
    EMPTY_PASSWORD: str = 'EMPTY_PASSWORD'
    USERNAME_ALREADY_EXISTS: str = 'USERNAME_ALREADY_EXISTS'
    EMAIL_ALREADY_EXISTS: str = 'EMAIL_ALREADY_EXISTS'
    INVALID_EMAIL: str = 'INVALID_EMAIL'
    INVALID_LENGTH_OF_USERNAME: str = 'INVALID_LENGTH_OF_USERNAME'


class TokenType:
    """Auth token type"""
    ACCESS_TOKEN = "access"
    REFRESH_TOKEN = "refresh"



"""
PLynx API for generation user Nodes.
"""
# pylint: disable=consider-using-from-import
from dataclasses import dataclass, field
from typing import Any, List, Optional

import cloudpickle

import plynx.node.typing as typings
import plynx.node.utils as utils  # noqa


@dataclass
class InputItem:
    """Input item abstraction"""
    name: str
    file_type: str
    is_array: bool
    min_count: int

    def to_dict(self):
        """Dict representation"""
        return {
            "name": self.name,
            "file_type": self.file_type,
            "is_array": self.is_array,
            "min_count": self.min_count,
        }


@dataclass
class OutputItem:
    """Output item abstraction"""
    name: str
    file_type: str
    is_array: bool
    min_count: int

    def to_dict(self):
        """Dict representation"""
        return {
            "name": self.name,
            "file_type": self.file_type,
            "is_array": self.is_array,
            "min_count": self.min_count,
        }


@dataclass
class ParamItem:
    """Parameter item abstraction"""
    name: str
    parameter_type: str
    value: Any
    widget: Optional[str]

    def to_dict(self):
        """Dict representation"""
        return {
            "name": self.name,
            "parameter_type": self.parameter_type,
            "value": self.value,
            "widget": self.widget,
        }


@dataclass
class PlynxParams:
    """Internal PLynx Node params"""
    # pylint: disable=too-many-instance-attributes
    title: str
    description: str
    kind: str
    node_type: str
    auto_run_enabled: bool = True
    inputs: List[InputItem] = field(default_factory=list)
    params: List[ParamItem] = field(default_factory=list)
    outputs: List[OutputItem] = field(default_factory=list)


# pylint: disable=redefined-builtin
def input(
    name=None,
    var_type=None,
    is_array=False,
    min_count=1,
):
    """PLynx Operation Input"""
    def decorator(func_or_class):
        func_or_class.plynx_params.inputs.insert(0, InputItem(
            name=name,
            file_type=typings.type_to_str(var_type),
            is_array=is_array,
            min_count=min_count,
        ))
        return func_or_class
    return decorator


def output(
    name=None,
    var_type=None,
    is_array=False,
    min_count=1,
):
    """PLynx Operation Output"""
    def decorator(func_or_class):
        func_or_class.plynx_params.outputs.insert(0, OutputItem(
            name=name,
            file_type=typings.type_to_str(var_type),
            is_array=is_array,
            min_count=min_count,
        ))
        return func_or_class
    return decorator


def param(
    name=None,
    var_type=None,
    default=None,
    widget="",
):
    """PLynx Operation Parameter"""
    def decorator(func_or_class):
        func_or_class.plynx_params.params.insert(0, ParamItem(
            name=name,
            parameter_type=typings.type_to_str(var_type),
            value=default,
            widget=None if widget is None else widget or name,
        ))
        return func_or_class
    return decorator


def operation(node_type=None, title=None, description="", kind=None, auto_run_enabled: bool = True):
    """PLynx user-defined Operation"""
    def decorator(func_or_class):
        func_or_class.plynx_params = PlynxParams(
            title=title or func_or_class.__name__,
            description=description,
            kind=kind or "python-code-operation",
            node_type=node_type or "python-code-operation",
            auto_run_enabled=auto_run_enabled,
            inputs=[],
            params=[],
            outputs=[],
        )

        fn_dumped = cloudpickle.dumps(func_or_class).hex()
        func_or_class.plynx_params.params.append(ParamItem(
            name="_pickled_fn",
            parameter_type="str",
            value=fn_dumped,
            widget=None,
        ))
        return func_or_class
    return decorator


parameter = param   # Alias



"""Cancelation DB Object and utils"""
from dataclasses import dataclass, field
from typing import List, Optional

from dataclasses_json import dataclass_json

from plynx.constants import Collections
from plynx.db.db_object import DBObject
from plynx.utils.common import ObjectId
from plynx.utils.db_connector import get_db_connector


@dataclass_json
@dataclass
class RunCancellation(DBObject):
    """RunCancellation represents Run Cancellation event in the database."""
    DB_COLLECTION = Collections.RUN_CANCELLATIONS

    _id: ObjectId = field(default_factory=ObjectId)
    run_id: Optional[ObjectId] = None


class RunCancellationManager:
    """RunCancellationManager contains basic operations related to `runs_cancellations` collection."""

    run_id: ObjectId

    @staticmethod
    def cancel_run(run_id: ObjectId):
        """Cancel Run.
        Args:
            run_id    (ObjectId) RunID
        """
        run_cancellation = RunCancellation(
            run_id=run_id,
        )
        run_cancellation.save()
        return True

    @staticmethod
    def get_run_cancellations() -> List[RunCancellation]:
        """Get all Run Cancellation events"""
        res = []
        for runs_cancellation_dict in get_db_connector()[Collections.RUN_CANCELLATIONS].find():
            res.append(
                RunCancellation.from_dict(runs_cancellation_dict)
            )
        return res

    @staticmethod
    def remove(runs_cancellation_ids: List[ObjectId]):
        """Remove Run Cancellation events with given Ids
        Args:
            runs_cancellation_ids     (list of ObjectID)  List of Run IDs to remove
        """
        get_db_connector()[Collections.RUN_CANCELLATIONS].delete_many({'_id': {'$in': runs_cancellation_ids}})



"""Parameter enums"""


class ParameterTypes:
    """Standard parameter types"""
    STR: str = 'str'
    INT: str = 'int'
    FLOAT: str = 'float'
    BOOL: str = 'bool'
    TEXT: str = 'text'
    ENUM: str = 'enum'
    LIST_STR: str = 'list_str'
    LIST_INT: str = 'list_int'
    LIST_NODE: str = 'list_node'
    CODE: str = 'code'



"""Global Node Registry class"""
import logging
from typing import Dict, List

from plynx.db.node import Node


class Registry:
    """The class keeps record of built-in Operations"""
    def __init__(self):
        self.code_function_location_to_node: Dict[str, Node] = {}
        self.code_function_location_to_hash: Dict[str, str] = {}

    def register_node(self, node: Node):
        """Register Node globally"""
        assert node.code_function_location, "`code_function_location` is not defined. Is this node from static list?"
        self.code_function_location_to_node[node.code_function_location] = node
        self.code_function_location_to_hash[node.code_function_location] = node.code_hash
        logging.warning(f"Registered operation with location `{node.code_function_location}`")

    def find_nodes(self, function_locations: List[str]) -> List[Node]:
        """Find the Nodes"""
        res = []
        for f_location in function_locations:
            if f_location not in self.code_function_location_to_node:
                continue
            res.append(self.code_function_location_to_node[f_location])
        return res


registry = Registry()



"""Commonly used Resource types and templates in python."""
import json
import os
import uuid
from typing import Any, Optional

import plynx.utils.file_handler
from plynx.base import resource


class Json(resource.BaseResource):
    """JSON file"""
    DISPLAY_THUMBNAIL: bool = True

    @staticmethod
    def preprocess_input(value: Any) -> Any:
        """Resource_id to an object"""

        with plynx.utils.file_handler.open(value, "r") as f:
            return json.load(f)

    @staticmethod
    def postprocess_output(value: Any) -> Any:
        """Object to resource id"""
        filename = str(uuid.uuid4())
        with plynx.utils.file_handler.open(filename, "w") as f:
            json.dump(value, f)
        return filename

    @classmethod
    def preview(cls, preview_object: resource.PreviewObject) -> str:
        """Generate preview html body"""
        res = json.load(preview_object.fp)
        return f"<pre>{json.dumps(res, indent=4)}</pre>"

    @classmethod
    def thumbnail(cls, output: Any) -> Optional[str]:
        if len(output.values) != 1:
            return None

        with plynx.utils.file_handler.open(output.values[0], "r") as f:
            res = json.load(f)
        formatted_text = json.dumps(res, indent=4)
        thumbnail_text = os.linesep.join(formatted_text.split(os.linesep)[:5])
        return f"<pre style='text-align:left; background:#222; margin:0; width:100%'>{thumbnail_text}...</pre>"



"""Main PLynx executor service and utils"""

import json

from plynx.utils import executor


def run_execute(filename: str):
    """Execute entrypoint. It materialize the Node based on file content and runs it."""
    with open(filename) as f:
        node = executor.materialize_executor(json.load(f))
    node.run()
    print('Success')
    return 0



"""
Service endpoints.
"""
# flake8: noqa
from gevent.pywsgi import WSGIServer

import plynx.web.health
import plynx.web.node
import plynx.web.resource
import plynx.web.state
import plynx.web.user
from plynx.web.common import app, authenticate, make_fail_response, register_user, requires_auth, run_api, verify_password



"""Node constants"""
from typing import Set

from bson.objectid import ObjectId


class NodeRunningStatus:
    """Running status enum"""
    STATIC: str = 'STATIC'
    CREATED: str = 'CREATED'
    READY: str = 'READY'
    IN_QUEUE: str = 'IN_QUEUE'
    RUNNING: str = 'RUNNING'
    SUCCESS: str = 'SUCCESS'
    RESTORED: str = 'RESTORED'
    FAILED: str = 'FAILED'
    FAILED_WAITING: str = 'FAILED_WAITING'
    CANCELED: str = 'CANCELED'
    SPECIAL: str = 'SPECIAL'

    _FAILED_STATUSES: Set[str] = {
        FAILED,
        CANCELED,
        FAILED_WAITING,
    }

    _SUCCEEDED_STATUSES: Set[str] = {
        STATIC,
        SUCCESS,
        RESTORED,
        SPECIAL,
    }

    _AWAITING_STATUSES: Set[str] = {
        READY,
        IN_QUEUE,
        RUNNING,
    }

    _FINISHED_STATUSES: Set[str] = _FAILED_STATUSES | _SUCCEEDED_STATUSES

    @staticmethod
    def is_finished(node_running_status: str) -> bool:
        """Check if the status is final"""
        return node_running_status in NodeRunningStatus._FINISHED_STATUSES

    @staticmethod
    def is_succeeded(node_running_status: str) -> bool:
        """Check if the status is final and successful"""
        return node_running_status in NodeRunningStatus._SUCCEEDED_STATUSES

    @staticmethod
    def is_failed(node_running_status: str) -> bool:
        """Check if the status is final and failed"""
        return node_running_status in NodeRunningStatus._FAILED_STATUSES


class NodeStatus:
    """Node permanent status"""
    CREATED: str = 'CREATED'
    READY: str = 'READY'
    DEPRECATED: str = 'DEPRECATED'
    MANDATORY_DEPRECATED: str = 'MANDATORY_DEPRECATED'


class NodePostAction:
    """HTTP post action"""
    SAVE: str = 'SAVE'
    APPROVE: str = 'APPROVE'
    CREATE_RUN: str = 'CREATE_RUN'
    CREATE_RUN_FROM_SCRATCH: str = 'CREATE_RUN_FROM_SCRATCH'
    CLONE: str = 'CLONE'
    VALIDATE: str = 'VALIDATE'
    DEPRECATE: str = 'DEPRECATE'
    MANDATORY_DEPRECATE: str = 'MANDATORY_DEPRECATE'
    PREVIEW_CMD: str = 'PREVIEW_CMD'
    REARRANGE_NODES: str = 'REARRANGE_NODES'
    UPGRADE_NODES: str = 'UPGRADE_NODES'
    CANCEL: str = 'CANCEL'
    GENERATE_CODE: str = 'GENERATE_CODE'


class NodePostStatus:
    """Standard HTTP response status"""
    SUCCESS: str = 'SUCCESS'
    FAILED: str = 'FAILED'
    VALIDATION_FAILED: str = 'VALIDATION_FAILED'


class NodeClonePolicy:
    """Clone algorithm"""
    NODE_TO_NODE: int = 0
    NODE_TO_RUN: int = 1
    RUN_TO_NODE: int = 2


class NodeVirtualCollection:
    """Virtual collection"""
    OPERATIONS: str = 'operations'
    WORKFLOWS: str = 'workflows'


class SpecialNodeId:
    """Special Node IDs in the workflows"""
    INPUT: ObjectId = ObjectId('2419f9500000000000000000')
    OUTPUT: ObjectId = ObjectId('56274ccc0000000000000000')


class NodeOrigin:
    """Enum that indicates where the Node came from"""
    DB: str = "DB"
    BUILT_IN_HUBS: str = "BUILT_IN_HUBS"


IGNORED_CACHE_PARAMETERS = {'cmd', '_timeout'}



"""DB connector"""
import logging

import pymongo

from plynx.constants import Collections
from plynx.utils.config import get_db_config

PLYNX_DB = "plynx"
_DB = None


def init_indexes():
    """Create DB indexes"""
    _DB[Collections.WORKER_STATE].create_index('insertion_date', expireAfterSeconds=5)

    _DB[Collections.RUNS].create_index('insertion_date')

    _DB[Collections.NODE_CACHE].create_index('key', unique=True)

    _DB[Collections.TEMPLATES].create_index('insertion_date')
    _DB[Collections.TEMPLATES].create_index([
        ('starred', pymongo.DESCENDING),
        ('insertion_date', pymongo.DESCENDING)
    ])
    _DB[Collections.TEMPLATES].create_index([('title', pymongo.TEXT), ('description', pymongo.TEXT)])

    _DB[Collections.RUNS].create_index('insertion_date')
    _DB[Collections.RUNS].create_index([('title', pymongo.TEXT), ('description', pymongo.TEXT)])

    _DB[Collections.USERS].create_index('username', unique=True)

    _DB[Collections.RUN_CANCELLATIONS].create_index('insertion_date', expireAfterSeconds=60)
    _DB[Collections.RUN_CANCELLATIONS].create_index('run_id')


def get_db_connector():
    """Create a connector lazily"""
    global _DB  # pylint: disable=global-statement
    if _DB is not None:
        return _DB

    connection_config, auth_params = get_db_config(), {}

    if connection_config.user and connection_config.password:
        auth_params.update({
            "username": connection_config.user,
            "password": connection_config.password,
            "authSource": PLYNX_DB
        })

    client = pymongo.MongoClient(
        connection_config.host,
        connection_config.port,
        read_preference=pymongo.read_preferences.PrimaryPreferred(),
        **auth_params
    )

    _DB = client[PLYNX_DB]

    init_indexes()
    return _DB


def check_connection():
    """Check DB connection"""
    try:
        logging.info('Try db connection')
        get_db_connector().client.server_info()
    except pymongo.errors.ServerSelectionTimeoutError:
        logging.error('Connection failed')
        raise



"""Executors supporting PLynx sync/async inference framework"""
import functools
import os
import shutil
import uuid

import plynx.db.run_cancellation_manager
from plynx.base.executor import BaseExecutor, RunningStatus
from plynx.constants import Collections
from plynx.db.node import Node


@functools.lru_cache()
def run_cancellation_manager():
    """Lazy RunCancellationManager object"""
    return plynx.db.run_cancellation_manager.RunCancellationManager()


class PLynxAsyncExecutor(BaseExecutor):
    """Base Executor class that is using PLynx Async Inference backend"""

    def launch(self) -> RunningStatus:
        """Put the Node on the queue"""
        assert self.node, "`node` in PLynxAsyncExecutor object is not defined"
        self.node.save(collection=Collections.RUNS)
        return RunningStatus(self.node.node_running_status)

    def kill(self):
        run_cancellation_manager().cancel_run(self.node._id)

    def get_running_status(self) -> RunningStatus:
        """Returns the status of the execution.

        Async executions should sync with the remote and return the result immediately.
        """
        assert self.node, "node must be defined at this point"
        self.node = Node.load(self.node._id, collection=Collections.RUNS)
        return RunningStatus(self.node.node_running_status)


class PLynxSyncExecutor(BaseExecutor):
    """Base Executor class that is using PLynx Sync Inference backend"""

    def launch(self) -> RunningStatus:
        """Run the node now and return the status"""
        assert self.node, "node must be defined at this point"
        self.run()
        return RunningStatus(self.node.node_running_status)


class PLynxAsyncExecutorWithDirectory(PLynxAsyncExecutor):
    """Base Executor class that is using PLynx Async Inference backend"""

    def __init__(self, node):
        super().__init__(node)
        self.workdir = os.path.join('/tmp', str(uuid.uuid1()))

    def init_executor(self):
        """Make tmp dir if it does not exist"""
        if not os.path.exists(self.workdir):
            os.makedirs(self.workdir)

    def clean_up_executor(self):
        """Remove tmp dir"""
        if os.path.exists(self.workdir):
            shutil.rmtree(self.workdir, ignore_errors=True)



"""Commonly used Resource types."""

import json
import os
import stat
import zipfile
from typing import Any, Dict, List, Optional

from plynx.base import resource
from plynx.constants import NodeResources
from plynx.utils.common import zipdir
from plynx.utils.config import WebConfig, get_web_config

WEB_CONFIG: WebConfig = get_web_config()


class Raw(resource.BaseResource):
    """Raw Resource that will be stored in jsonable format in the Node."""
    DISPLAY_RAW: bool = True


class File(resource.BaseResource):
    """Raw Resource that will be stored in the file format in the Node."""


class PDF(resource.BaseResource):
    """PDF file"""
    @classmethod
    def preview(cls, preview_object: resource.PreviewObject) -> str:
        """Generate preview html body"""
        src_url = f"{WEB_CONFIG.endpoint}/resource/{preview_object.resource_id}"
        return f'<iframe src="{src_url}" title="preview" type="application/pdf" width="100%"/>'


class Image(resource.BaseResource):
    """Image file"""
    DISPLAY_THUMBNAIL: bool = True

    @classmethod
    def preview(cls, preview_object: resource.PreviewObject) -> str:
        """Generate preview html body"""
        src_url = f"{WEB_CONFIG.endpoint}/resource/{preview_object.resource_id}"
        return f'<img src="{src_url}" width="100%" alt="preview" />'

    @classmethod
    def thumbnail(cls, output: Any) -> Optional[str]:
        if len(output.values) != 1:
            return None
        src_url = f"{WEB_CONFIG.endpoint}/resource/{output.values[0]}"
        return f'<img src="{src_url}" width="100%" alt="preview" />'


class _BaseSeparated(resource.BaseResource):
    """Base Separated file, i.e. csv, tsv"""
    SEPARATOR: Optional[str] = None
    _ROW_CLASSES: List[str] = ['even', 'odd']
    _NUM_ROW_CLASSES: int = len(_ROW_CLASSES)

    @classmethod
    def preview(cls, preview_object: resource.PreviewObject) -> str:
        """Generate preview html body"""
        preview_object.fp.truncate(1024 ** 2)
        formated_lines = []
        for idx, line in enumerate(preview_object.fp.read().decode('utf-8').split('\n')):
            even_or_odd = cls._ROW_CLASSES[idx % cls._NUM_ROW_CLASSES]
            formated_cells = ''.join(map(lambda s: f'<td class="preview-table-col">{s}</td>', line.split(cls.SEPARATOR)))
            formated_line = f'<tr class="preview-table-row-{even_or_odd}">{formated_cells}</tr>'
            formated_lines.append(formated_line)
        all_lines = '\n'.join(formated_lines)
        return f'<table class="preview-table">{all_lines}</table>'


class CSV(_BaseSeparated):
    """CSV file"""
    SEPARATOR: str = ','


class TSV(_BaseSeparated):
    """TSV file"""
    SEPARATOR: str = '\t'


class Json(resource.BaseResource):
    """JSON file"""
    @classmethod
    def preview(cls, preview_object: resource.PreviewObject) -> str:
        """Generate preview html body"""
        if preview_object.fp.getbuffer().nbytes > 1024 ** 2:
            return super(Json, cls).preview(preview_object)
        try:
            readable_json = json.dumps(json.loads(preview_object.fp.read().decode('utf-8')), indent=2)
            return f"<pre>{readable_json}</pre>"
        except json.JSONDecodeError as e:
            return f"Failed to parse json: {e}"


class Executable(resource.BaseResource):
    """Executable file, i.e. bash or python"""
    @staticmethod
    def prepare_input(filename, preview: bool = False) -> Dict[str, str]:
        """Generate preview html body"""
        # `chmod +x` to the executable file
        if preview:
            return {NodeResources.INPUT: filename}
        file_status = os.stat(filename)
        os.chmod(filename, file_status.st_mode | stat.S_IEXEC)
        return {NodeResources.INPUT: filename}


class Directory(resource.BaseResource):
    """Directory file, i.e. zipfile"""
    @staticmethod
    def prepare_input(filename, preview: bool = False) -> Dict[str, str]:
        """Extract zip file"""
        if preview:
            return {NodeResources.INPUT: filename}
        zip_filename = f"{filename}.zip"
        os.rename(filename, zip_filename)
        os.mkdir(filename)
        with zipfile.ZipFile(zip_filename) as zf:
            zf.extractall(filename)
        return {NodeResources.INPUT: filename}

    @staticmethod
    def prepare_output(filename, preview: bool = False) -> Dict[str, str]:
        """Create output folder"""
        if preview:
            return {NodeResources.OUTPUT: filename}
        os.mkdir(filename)
        return {NodeResources.OUTPUT: filename}

    @staticmethod
    def postprocess_output(value: str) -> str:
        """Compress folder to a zip file"""
        zip_filename = f"{value}.zip"
        with zipfile.ZipFile(zip_filename, 'w', zipfile.ZIP_DEFLATED) as zf:
            zipdir(value, zf)
        return zip_filename

    @classmethod
    def preview(cls, preview_object: resource.PreviewObject) -> str:
        """Generate preview html body"""
        with zipfile.ZipFile(preview_object.fp, 'r') as zf:
            content_stream = '\n'.join(zf.namelist())

        return f"<pre>{content_stream}</pre>"


FILE_KIND = 'file'



"""Templates for PLynx Executors and utils."""
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional, Union

from plynx.constants import NodeStatus, SpecialNodeId, ValidationCode, ValidationTargetType
from plynx.db.node import Node, NodeRunningStatus, Parameter, ParameterListOfNodes, ParameterTypes
from plynx.db.validation_error import ValidationError


@dataclass
class RunningStatus:
    """Async job running status"""
    node_running_status: str


class BaseExecutor(ABC):
    """Base Executor class"""
    IS_GRAPH: bool = False

    def __init__(self, node: Optional[Node] = None):
        self.node = node

    def _update_node(self, node):
        self.node = node

    @abstractmethod
    def run(self, preview: bool = False) -> str:
        """Main execution function.

        - Workdir has been initialized.
        - Inputs are not preprocessed.
        - Outputs shoul be manually postprocessed.
        - It is OK to raise an exception in this function.

        Returns:
            enum: plynx.constants.NodeRunningStatus
        """

    @abstractmethod
    def launch(self) -> RunningStatus:
        """Launch the Node on the backend.

        The difference between `launch()` and `run()` is that:
        - `run()` assumes synchronous execution.
        - `launch()` does not necessary have this assumtion. Use `get_node_running_status` to get the status of the execution.

        Returns:
            RunningStatus
        """
        raise NotImplementedError()

    def get_running_status(self) -> RunningStatus:
        """Returns the status of the execution.

        Async executions should sync with the remote and return the result immediately.
        """
        assert self.node, "Node must be difened by this stage"
        return RunningStatus(self.node.node_running_status)

    @abstractmethod
    def kill(self):
        """Force to kill the process.

        The reason can be the fact it was working too long or parent executor canceled it.
        """

    def is_updated(self) -> bool:
        """Function that is regularly called by a Worker.

        The function is running in a separate thread and does not block execution of `run()`.

        Returns:
            (bool):     True if worker needs to update DB else False
        """
        return False

    @classmethod
    def get_default_node(cls, is_workflow: bool) -> Node:
        """Generate a new default Node for this executor"""
        node = Node()
        if cls.IS_GRAPH:
            nodes_parameter = Parameter(
                name='_nodes',
                parameter_type=ParameterTypes.LIST_NODE,
                value=ParameterListOfNodes(),
                mutable_type=False,
                publicable=False,
                removable=False,
            )
            if not is_workflow:
                # need to add inputs and outputs
                nodes_parameter.value.value.extend(
                    [
                        Node(
                            _id=SpecialNodeId.INPUT,
                            title='Input',
                            kind='dummy',
                            node_running_status=NodeRunningStatus.SPECIAL,
                            node_status=NodeStatus.READY,
                        ),
                        Node(
                            _id=SpecialNodeId.OUTPUT,
                            title='Output',
                            kind='dummy',
                            node_running_status=NodeRunningStatus.SPECIAL,
                            node_status=NodeStatus.READY,
                        ),
                    ]
                )
            node.parameters.extend([
                nodes_parameter,
            ])
        return node

    def init_executor(self):
        """Initialize environment for the executor"""

    def clean_up_executor(self):
        """Clean up the environment created by executor"""

    def validate(self, ignore_inputs: bool = True) -> Union[ValidationError, None]:
        """Validate Node.

        Return:
            (ValidationError)   Validation error if found; else None
        """
        assert self.node, "Attribute `node` is not assigned"

        violations = []

        if self.node.title == '':
            violations.append(
                ValidationError(
                    target=ValidationTargetType.PROPERTY,
                    object_id='title',
                    validation_code=ValidationCode.MISSING_PARAMETER
                ))

        # Meaning the node is in the graph. Otherwise souldn't be in validation step
        if not ignore_inputs:
            for input in self.node.inputs:  # pylint: disable=redefined-builtin
                min_count = input.min_count if input.is_array else 1
                if len(input.input_references) < min_count:
                    violations.append(
                        ValidationError(
                            target=ValidationTargetType.INPUT,
                            object_id=input.name,
                            validation_code=ValidationCode.MISSING_INPUT
                        ))

        if self.node.node_status == NodeStatus.MANDATORY_DEPRECATED:
            violations.append(
                ValidationError(
                    target=ValidationTargetType.NODE,
                    object_id=str(self.node._id),
                    validation_code=ValidationCode.DEPRECATED_NODE
                ))

        if len(violations) == 0:
            return None

        return ValidationError(
            target=ValidationTargetType.NODE,
            object_id=str(self.node._id),
            validation_code=ValidationCode.IN_DEPENDENTS,
            children=violations
        )


class Dummy(BaseExecutor):
    """Dummy Executor. Used for static Operations"""

    def run(self, preview=False) -> str:
        """Not Implemented"""
        raise NotImplementedError()

    def status(self):
        """Not Implemented"""
        raise NotImplementedError()

    def kill(self):
        """Not Implemented"""
        raise NotImplementedError()

    @classmethod
    def get_default_node(cls, is_workflow: bool) -> Node:
        """Not Implemented"""
        raise NotImplementedError()



"""Node DB Object and utils"""
from dataclasses import dataclass, field
from typing import Any, Callable, List, Optional, Union

from dataclasses_json import dataclass_json
from past.builtins import basestring

from plynx.constants import Collections, NodeClonePolicy, NodeOrigin, NodeRunningStatus, NodeStatus, ParameterTypes
from plynx.db.db_object import DBObject
from plynx.plugins.resources.common import FILE_KIND
from plynx.utils.common import ObjectId


# pylint: disable=too-many-branches
def _clone_update_in_place(node: "Node", node_clone_policy: int, override_finished_state: bool, override_node_id: bool = False):
    if node.node_running_status == NodeRunningStatus.SPECIAL:
        for output in node.outputs:
            output.values = []
        return node
    old_node_id = node._id
    node.template_node_id = node._id
    if override_node_id:
        node._id = ObjectId()
    node.successor_node_id = None

    if node_clone_policy == NodeClonePolicy.NODE_TO_NODE:
        node.parent_node_id = old_node_id
        node.original_node_id = None
    elif node_clone_policy == NodeClonePolicy.NODE_TO_RUN:
        node.parent_node_id = None
        node.original_node_id = node.original_node_id or old_node_id
    elif node_clone_policy == NodeClonePolicy.RUN_TO_NODE:
        node.parent_node_id = node.original_node_id
        node.original_node_id = None
    else:
        raise Exception(f"Unknown clone policy `{node_clone_policy}`")

    if node.node_running_status == NodeRunningStatus.STATIC:
        return node

    if override_finished_state or not NodeRunningStatus.is_succeeded(node.node_running_status):
        node.node_running_status = NodeRunningStatus.READY
    node.node_status = NodeStatus.CREATED

    sub_nodes = node.get_parameter_by_name('_nodes', throw=False)
    if sub_nodes:
        object_id_mapping = {}
        for sub_node in sub_nodes.value.value:
            prev_id = ObjectId(sub_node._id)
            _clone_update_in_place(sub_node, node_clone_policy, override_finished_state)
            object_id_mapping[prev_id] = sub_node._id

        for sub_node in sub_nodes.value.value:
            for input in sub_node.inputs:   # pylint: disable=redefined-builtin
                input.values = []
                for input_reference in input.input_references:
                    input_reference.node_id = object_id_mapping[ObjectId(input_reference.node_id)]

            for parameter in sub_node.parameters:
                if not parameter.reference:
                    continue
                parameter.value = node.get_parameter_by_name(parameter.reference, throw=True).value

            if sub_node.node_running_status == NodeRunningStatus.STATIC:
                # do not copy the rest of the elements because they don't change
                continue

            if override_finished_state or not NodeRunningStatus.is_succeeded(sub_node.node_running_status):
                for output in sub_node.outputs:
                    output.values = []

                for log in sub_node.logs:
                    log.values = []

    if override_finished_state or not NodeRunningStatus.is_succeeded(node.node_running_status):
        for output_or_log in node.outputs + node.logs:
            output_or_log.values = []
    return node


@dataclass_json
@dataclass
class _BaseResource(DBObject):
    name: str = ""
    file_type: str = FILE_KIND
    values: List[Any] = field(default_factory=list)
    is_array: bool = False
    min_count: int = 1


@dataclass_json
@dataclass
class Output(_BaseResource):
    """Basic Output structure."""

    thumbnail: Optional[str] = None


@dataclass_json
@dataclass
class InputReference(DBObject):
    """Basic Value of the Input structure."""

    node_id: str = ""
    output_id: str = ""


@dataclass_json
@dataclass
class Input(_BaseResource):
    """Basic Input structure."""

    input_references: List[InputReference] = field(default_factory=list)


@dataclass_json
@dataclass
class CachedNode(DBObject):
    """Values to override Node on display"""
    node_running_status: str = NodeRunningStatus.CREATED
    outputs: List[Output] = field(default_factory=list)
    logs: List[Output] = field(default_factory=list)


@dataclass_json
@dataclass
class Node(DBObject):
    """Basic Node with db interface."""
    # pylint: disable=too-many-instance-attributes
    DB_COLLECTION = Collections.TEMPLATES

    _id: ObjectId = field(default_factory=ObjectId)
    _type: str = "Node"
    _cached_node: Optional[CachedNode] = None
    title: str = "Title"
    description: str = ""
    kind: str = "dummy"
    # ID of previous version of the node, always refer to `nodes` collection.
    parent_node_id: Optional[ObjectId] = None
    # ID of next version of the node, always refer to `nodes` collection.
    successor_node_id: Optional[ObjectId] = None
    # ID of original node, used in `runs`, always refer to `nodes` collection.
    # A Run refers to original node
    original_node_id: Optional[ObjectId] = None
    template_node_id: Optional[ObjectId] = None
    origin: str = NodeOrigin.DB
    # The following `code_*` values defined when the operation declared outside of plynx DB
    # code_hash is a hash value of the code
    code_hash: str = ""
    # code_function_location refers to the location of the function
    code_function_location: Optional[str] = None
    node_running_status: str = NodeRunningStatus.CREATED
    node_status: str = NodeStatus.CREATED
    cache_url: Optional[str] = None
    x: int = 0
    y: int = 0
    author: Optional[ObjectId] = None
    starred: bool = False
    auto_sync: bool = True
    auto_run: bool = True
    auto_run_enabled: bool = True
    # latest_run_id would be used in Templates to refer to the current run
    latest_run_id: Optional[ObjectId] = None

    inputs: List[Input] = field(default_factory=list)
    parameters: List["Parameter"] = field(default_factory=list)
    outputs: List[Output] = field(default_factory=list)
    logs: List[Output] = field(default_factory=list)

    @staticmethod
    def _default_log(name: str) -> Output:
        return Output(
            name=name,
            file_type=FILE_KIND,
            values=[],
            is_array=False,
            min_count=1,
        )

    # pylint: disable=attribute-defined-outside-init
    def apply_properties(self, other_node: "Node"):
        """Apply Properties and Inputs of another Node.
        This method is used for updating nodes.

        Args:
            other_node  (Node):     A node to copy Properties and Inputs from
        """
        for other_input in other_node.inputs:
            for input in self.inputs:   # pylint: disable=redefined-builtin
                if other_input.name == input.name and \
                   other_input.file_type == input.file_type and (
                            input.is_array or
                            (not input.is_array and 1 == len(other_input.input_references))
                       ):
                    input.input_references = other_input.input_references
                    break

        for other_parameter in other_node.parameters:
            for parameter in self.parameters:
                if other_parameter.name == parameter.name:
                    if parameter.parameter_type == other_parameter.parameter_type and parameter.widget:
                        parameter.value = other_parameter.value
                    break

        self.description = other_node.description

        self.x = other_node.x
        self.y = other_node.y

    def clone(self, node_clone_policy: int, override_finished_state: bool = True) -> "Node":
        """Return a cloned copy of a Node"""
        node = _clone_update_in_place(Node.from_dict(self.to_dict()), node_clone_policy, override_finished_state, override_node_id=True)
        return node

    def _get_custom_element(
                self,
                arr: Union[List[Input], List["Parameter"], List[Output]],
                name: str,
                throw: bool,
                default: Optional[Callable[[str], Union[Input, "Parameter", Output]]] = None
            ):
        for obj in arr:
            if obj.name == name:
                return obj
        if throw:
            raise Exception(f'Object "{name}" not found in {self.title}')
        if default:
            arr.append(default(name=name))  # type: ignore
            return arr[-1]
        return None

    def get_input_by_name(self, name: str, throw: bool = True):
        """Find Input object"""
        return self._get_custom_element(self.inputs, name, throw)

    def get_parameter_by_name(self, name: str, throw: bool = True):
        """Find Parameter object"""
        return self._get_custom_element(self.parameters, name, throw)

    def get_output_by_name(self, name: str, throw: bool = True):
        """Find Output object"""
        return self._get_custom_element(self.outputs, name, throw)

    def get_log_by_name(self, name: str, throw: bool = False):
        """Find Log object"""
        return self._get_custom_element(self.logs, name, throw, default=Node._default_log)

    def get_sub_nodes(self):
        """Get a list of subnodes"""
        sub_nodes_parameter = self.get_parameter_by_name('_nodes', throw=False)
        if not sub_nodes_parameter:
            raise Exception("Subnodes not found")
        return sub_nodes_parameter.value.value


@dataclass_json
@dataclass
class ParameterEnum(DBObject):
    """Enum value."""
    values: List[str] = field(default_factory=list)
    index: int = 0


@dataclass_json
@dataclass
class ParameterCode(DBObject):
    """Code value."""
    value: str = ""
    mode: str = "python"


@dataclass_json
@dataclass
class ParameterListOfNodes(DBObject):
    """List Of Nodes value."""
    value: List[Node] = field(default_factory=list)


# pylint: disable=too-many-return-statements
def _get_default_by_type(parameter_type: str):
    if parameter_type == ParameterTypes.STR:
        return ''
    if parameter_type == ParameterTypes.INT:
        return 0
    if parameter_type == ParameterTypes.FLOAT:
        return 0
    if parameter_type == ParameterTypes.BOOL:
        return False
    if parameter_type == ParameterTypes.TEXT:
        return ''
    if parameter_type == ParameterTypes.ENUM:
        return ParameterEnum()
    if parameter_type == ParameterTypes.LIST_STR:
        return []
    if parameter_type == ParameterTypes.LIST_INT:
        return []
    elif parameter_type == ParameterTypes.LIST_NODE:
        return ParameterListOfNodes()
    elif parameter_type == ParameterTypes.CODE:
        return ParameterCode()
    else:
        return None


def _value_is_valid(value, parameter_type: str):
    if parameter_type == ParameterTypes.STR:
        return isinstance(value, basestring)
    if parameter_type == ParameterTypes.INT:
        try:
            int(value)
        except ValueError:
            return False
        return True
    if parameter_type == ParameterTypes.FLOAT:
        try:
            float(str(value))
        except ValueError:
            return False
        return True
    if parameter_type == ParameterTypes.BOOL:
        return isinstance(value, int)
    if parameter_type == ParameterTypes.TEXT:
        return isinstance(value, basestring)
    if parameter_type == ParameterTypes.ENUM:
        return isinstance(value, ParameterEnum)
    if parameter_type == ParameterTypes.LIST_STR:
        return isinstance(value, list) and all(_value_is_valid(x, ParameterTypes.STR) for x in value)
    if parameter_type == ParameterTypes.LIST_INT:
        return isinstance(value, list) and all(_value_is_valid(x, ParameterTypes.INT) for x in value)
    elif parameter_type == ParameterTypes.LIST_NODE:
        # TODO proper validation
        return isinstance(value, ParameterListOfNodes)
    if parameter_type == ParameterTypes.CODE:
        return isinstance(value, ParameterCode)
    else:
        return False


@dataclass_json
@dataclass
class Parameter(DBObject):
    """Basic Parameter structure."""
    # pylint: disable=too-many-instance-attributes

    name: str = ""
    parameter_type: str = ParameterTypes.STR
    # TODO make type factory
    value: Any = ""
    mutable_type: bool = True
    removable: bool = True
    publicable: bool = True
    widget: Optional[str] = None
    reference: Optional[str] = None

    def __post_init__(self):
        # `value` field is a special case: the type depends on `parameter_type`
        if self.value is None:
            self.value = _get_default_by_type(self.parameter_type)
        elif self.parameter_type == ParameterTypes.ENUM:
            self.value = ParameterEnum.from_dict(self.value)
        elif self.parameter_type == ParameterTypes.CODE:
            self.value = ParameterCode.from_dict(self.value)
        elif self.parameter_type == ParameterTypes.LIST_NODE:
            self.value = ParameterListOfNodes.from_dict(self.value)
        if not _value_is_valid(self.value, self.parameter_type):
            raise ValueError(f"Invalid parameter value type: {self.name}: {self.value}")



"""Templates for PLynx Hubs and utils."""

from abc import abstractmethod
from dataclasses import dataclass
from typing import Any, Dict, Optional

from dataclasses_json import dataclass_json


@dataclass_json
@dataclass
class Query:
    """Hub search query entry"""
    status: str = ""
    search: str = ""
    per_page: int = 30
    offset: int = 0
    user_id: Optional[str] = None


class BaseHub:
    """Base Hub class"""
    def __init__(self):
        pass

    @abstractmethod
    def search(self, query: Query) -> Dict[str, Any]:
        """Search for items given a query"""



"""Node validation enums"""


class ValidationTargetType:
    """Object Target"""
    BLOCK: str = 'BLOCK'
    GRAPH: str = 'GRAPH'
    INPUT: str = 'INPUT'
    NODE: str = 'NODE'
    PARAMETER: str = 'PARAMETER'
    PROPERTY: str = 'PROPERTY'


class ValidationCode:
    """Standard validation code"""
    IN_DEPENDENTS: str = 'IN_DEPENDENTS'
    MISSING_INPUT: str = 'MISSING_INPUT'
    MISSING_PARAMETER: str = 'MISSING_PARAMETER'
    INVALID_VALUE: str = 'INVALID_VALUE'
    DEPRECATED_NODE: str = 'DEPRECATED_NODE'
    EMPTY_GRAPH: str = 'EMPTY_GRAPH'



"""Main PLynx users service and utils"""
from typing import Optional

from plynx.db.user import User, UserCollectionManager

LIST_USERS = 'list_users'
CREATE_USER = 'create_user'
ACTIVATE_USER = 'activate_user'
DEACTIVATE_USER = 'deactivate_user'
MODES = [
    LIST_USERS,
    CREATE_USER,
    ACTIVATE_USER,
    DEACTIVATE_USER,
]


def run_list_users() -> None:
    """List all users"""
    for user_dict in User.find_users():
        user = User.from_dict(user_dict)
        print(','.join(map(str, [user._id, user.username])))


def run_create_user(email: Optional[str], username: Optional[str], password: Optional[str]) -> User:
    """Create a user"""
    if not username:
        raise ValueError('Username must be specified')
    password = password or ''
    user = User()
    user.username = username
    user.email = email
    user.hash_password(password)
    user.save()
    print(f"User `{username}` created")
    return user


def run_set_activation(username: Optional[str], value: bool) -> None:
    """Set user active status"""
    assert username, "Argument `username` is undifined"
    user = UserCollectionManager.find_user_by_name(username)

    if not user:
        raise ValueError(f"Username `{username}` not found")

    user.active = value
    user.save()
    print(f"User`s `{username}` active state changed to {value}")


def run_users(mode: str, email: Optional[str] = None, username: Optional[str] = None, password: Optional[str] = ''):
    """Users CLI entrypoint"""
    if mode not in MODES:
        raise ValueError(f"`mode` must be one of `{MODES}`. Value `{mode}` is given")
    if mode == LIST_USERS:
        run_list_users()
    elif mode == CREATE_USER:
        run_create_user(email, username, password)
    elif mode == ACTIVATE_USER:
        run_set_activation(username, True)
    elif mode == DEACTIVATE_USER:
        run_set_activation(username, False)



"""Node Cache and utils"""

import hashlib
from builtins import str
from dataclasses import dataclass, field
from typing import List, Optional

from dataclasses_json import dataclass_json

from plynx.constants import IGNORED_CACHE_PARAMETERS, Collections
from plynx.db.db_object import DBObject
from plynx.db.node import Node, Output
from plynx.utils.common import ObjectId
from plynx.utils.config import DemoConfig, get_demo_config

demo_config: DemoConfig = get_demo_config()


@dataclass_json
@dataclass
class NodeCache(DBObject):
    """Basic Node Cache with db interface."""
    # pylint: disable=too-many-instance-attributes
    DB_COLLECTION = Collections.NODE_CACHE

    _id: ObjectId = field(default_factory=ObjectId)
    key: str = ""
    node_id: Optional[ObjectId] = None
    run_id: Optional[ObjectId] = None
    outputs: List[Output] = field(default_factory=list)
    logs: List[Output] = field(default_factory=list)
    # TODO check if those are working
    removed: bool = False
    protected: bool = False

    @staticmethod
    def instantiate(node: Node, run_id: ObjectId) -> "NodeCache":
        """Instantiate a Node Cache from Node.

        Args:
            node    (Node):     Node object
            run_id  (ObjectId): Run ID

        Return:
            (NodeCache)
        """

        return NodeCache(
            key=NodeCache.generate_key(node),
            node_id=node._id,
            run_id=run_id,
            outputs=node.outputs,
            logs=node.logs,
        )

    @staticmethod
    def generate_key(node: Node) -> str:
        """Generate hash.

        Args:
            node    (Node): Node object

        Return:
            (str)   Hash value
        """
        inputs = node.inputs
        parameters = node.parameters
        original_node_id = node.original_node_id

        sorted_inputs = sorted(inputs, key=lambda x: x.name)
        inputs_hash = ','.join([
            f"{input.name}:{','.join(sorted(map(str, input.values)))}"
            for input in sorted_inputs
        ])

        sorted_parameters = sorted(parameters, key=lambda x: x.name)
        parameters_hash = ','.join([
            f"{parameter.name}:{parameter.value}"
            for parameter in sorted_parameters if parameter.name not in IGNORED_CACHE_PARAMETERS
        ])

        return hashlib.sha256(
            ';'.join([
                    str(original_node_id),
                    inputs_hash,
                    parameters_hash,
                ]).encode('utf-8')
        ).hexdigest()



"""Standard PLynx collections in DB"""


class Collections:
    """All basic collections used in DB"""
    NODE_CACHE: str = 'node_cache'
    RUN_CANCELLATIONS: str = 'run_cancellations'
    RUNS: str = 'runs'
    TEMPLATES: str = 'templates'
    USERS: str = 'users'
    WORKER_STATE: str = 'worker_state'

    # virtual collections:
    HUB_NODE_REGISTRY: str = 'hub_node_registry'



"""Worker State DB Object and utils"""

from dataclasses import dataclass, field
from typing import List

from dataclasses_json import dataclass_json

from plynx.constants import Collections
from plynx.db.db_object import DBObject
from plynx.db.node import Node
from plynx.utils.common import ObjectId
from plynx.utils.db_connector import get_db_connector


@dataclass_json
@dataclass
class WorkerState(DBObject):
    """Worker statistics snapshot."""
    DB_COLLECTION = Collections.WORKER_STATE

    _id: ObjectId = field(default_factory=ObjectId)
    worker_id: str = ""
    host: str = ""
    runs: List[Node] = field(default_factory=list)
    kinds: List[str] = field(default_factory=list)


def get_worker_states() -> List[WorkerState]:
    """Get all of the workers with latest states"""
    states = getattr(get_db_connector(), Collections.WORKER_STATE).find({}).sort('insertion_date', -1)

    unique_worker_states = {}
    for state_dict in states:
        if state_dict['worker_id'] in unique_worker_states:
            continue
        unique_worker_states[state_dict['worker_id']] = WorkerState.from_dict(state_dict)

    return list(sorted(unique_worker_states.values(), key=lambda state: state.worker_id))



"""Plynx standard Hub based on the database of Operations"""
from typing import Any, Dict

from plynx.base import hub
from plynx.db.node_collection_manager import NodeCollectionManager


class CollectionHub(hub.BaseHub):
    """Plynx standard Hub based on the database of Operations"""
    def __init__(self, collection, operations):
        super().__init__()

        self.node_collection_manager = NodeCollectionManager(collection=collection)
        self.operations = operations

    def search(self, query: hub.Query) -> Dict[str, Any]:
        return self.node_collection_manager.get_db_objects(
            node_kinds=self.operations,
            **query.to_dict(),  # type: ignore
        )



"""Standard PLynx types"""
from typing import Callable, Dict, Type, Union

ANY = "any"

STR_TO_CLASS: Dict[str, Union[Type, Callable]] = {
    "int": int,
    "str": str,
    "dict": dict,
    "float": float,
    "bool": bool,
    # Any other
    ANY: lambda x: x,
}

CLASS_TO_STR: Dict[Union[Type, Callable], str] = {
    v: k for k, v in STR_TO_CLASS.items()
}


def type_to_str(type_cls: Union[str, Type, Callable]) -> str:
    """Standard type to PLynx type"""
    if isinstance(type_cls, str):
        return type_cls
    return CLASS_TO_STR[type_cls]



"""Endpoints responsible for the dashboard"""
import plynx.utils.plugin_manager
from plynx.db.worker_state import get_worker_states
from plynx.web.common import app, handle_errors, logger, make_fail_response, make_success_response, requires_auth

PLUGINS_DICT = plynx.utils.plugin_manager.get_plugins_dict()


@app.route('/plynx/api/v0/worker_states', methods=['GET', 'POST'])
@handle_errors
@requires_auth
def worker_states():
    """Get worker's states"""
    try:
        return make_success_response({
            'items': list(map(lambda worker_state: worker_state.to_dict(), get_worker_states())),
            'plugins_dict': PLUGINS_DICT,
        })
    except Exception as e:  # pylint: disable=broad-except
        logger.error(e)
        return make_fail_response(f'Internal error: "{e}"')



"""An executor for the DAGs based on python backend."""
import logging
import multiprocessing
import queue
import traceback
import uuid
from typing import List

import plynx.plugins.executors.dag
import plynx.utils.executor
from plynx.constants import Collections, NodeRunningStatus
from plynx.db.node import Node
from plynx.utils import node_utils
from plynx.utils.common import to_object_id

POOL_SIZE = 3


def worker_main(job_run_queue: queue.Queue, job_complete_queue: queue.Queue):
    """Main threaded function that serves Operations."""
    logging.info("Created pool worker")
    while True:
        node = job_run_queue.get()
        executor = plynx.utils.executor.materialize_executor(node)

        try:
            # TODO support async executors
            executor.launch()
        except Exception:   # pylint: disable=broad-except
            error_str = traceback.format_exc()
            logging.error(f"Job failed with traceback: {error_str}")
            node.node_running_status = NodeRunningStatus.FAILED

            err_filename = str(uuid.uuid4())
            with plynx.utils.file_handler.open(err_filename, "w") as f:
                f.write(error_str)
            output = node.get_log_by_name(name="traceback")
            output.values = [err_filename]
            logging.error(f"Wrote to logs: {err_filename}")
        else:
            node.node_running_status = NodeRunningStatus.SUCCESS
        job_complete_queue.put(node)
        logging.info(f"worker_main: added node {node.title} on the job_complete_queue queue")

    logging.info("Deleted pool worker")


class DAGParallel(plynx.plugins.executors.dag.DAG):
    """ Python DAG scheduler.

    Args:
        node_dict (dict)
    """
    IS_GRAPH = True
    GRAPH_ITERATION_SLEEP = 0

    def __init__(self, node: Node):
        super().__init__(node)

        self.job_run_queue: queue.Queue = queue.Queue()
        self.job_complete_queue: queue.Queue = queue.Queue()

        self.worker_pool = multiprocessing.pool.ThreadPool(
            POOL_SIZE, worker_main, (
                self.job_run_queue,
                self.job_complete_queue,
            )
        )

    def pop_jobs(self) -> List[Node]:
        """Get a set of nodes with satisfied dependencies"""
        assert self.node, "Attribute `node` is undefined"
        res: List[Node] = []

        num_completed_jobs = 0
        while not self.job_complete_queue.empty():
            try:
                node = self.job_complete_queue.get_nowait()
                logging.info(f"pop_jobs: run update {node.title}: start")
                self.update_node(node)
                logging.info(f"pop_jobs: run update {node.title}: finish")
                num_completed_jobs += 1
            except queue.Empty:
                logging.warning("Queue is empty")
                break

        if NodeRunningStatus.is_failed(self._node_running_status):
            if self._node_running_status != NodeRunningStatus.FAILED_WAITING:
                logging.info(f"Job in DAG failed with status {self._node_running_status}, pop_jobs will return []")
            return res

        for node_id in self.dependency_index_to_node_ids[0]:
            # Get the node and init its inputs, i.e. filling its resource_ids
            node = self.node_id_to_node[node_id]
            for node_input in node.inputs:
                for input_reference in node_input.input_references:
                    node_input.values.extend(
                        self.node_id_to_node[to_object_id(input_reference.node_id)].get_output_by_name(
                            input_reference.output_id
                        ).values
                    )
            node.node_running_status = NodeRunningStatus.IN_QUEUE

            res.append(node)
        del self.dependency_index_to_node_ids[0]

        if num_completed_jobs > 0:
            self.node.save(collection=Collections.RUNS, force=True)

        return res

    def _execute_node(self, node: Node):
        assert self.node, "Attribute `node` is undefined"
        if NodeRunningStatus.is_finished(node.node_running_status):     # NodeRunningStatus.SPECIAL
            return

        logging.info(f'Execute {node} {node.title}')
        node.node_running_status = NodeRunningStatus.RUNNING
        self.node.save(collection=Collections.RUNS)
        # TODO somehow optimize `update_node`?
        # If not copy but original sent, the dependencies list won't be updated
        self.job_run_queue.put(node.copy())

    def kill(self):
        """Force to kill the process.

        The reason can be the fact it was working too long or parent exectuter canceled it.
        """
        logging.info("Received kill request")
        self._node_running_status = NodeRunningStatus.CANCELED
        self.worker_pool.terminate()

    def finished(self) -> bool:
        """Return True or False depending on the running status of the DAG."""
        # TODO wait for the canceled nodes to complete
        return self._node_running_status in {NodeRunningStatus.SUCCESS, NodeRunningStatus.FAILED, NodeRunningStatus.CANCELED}


class DAG(plynx.plugins.executors.dag.DAG):
    """Base Executor class"""
    IS_GRAPH: bool = True

    def __init__(self, node: Node):
        super().__init__(node)
        self.job_run_queue: queue.Queue = queue.Queue()
        self.job_complete_queue: queue.Queue = queue.Queue()
        self.worker_pool = None

    def kill(self):
        """Force to kill the process.

        The reason can be the fact it was working too long or parent exectuter canceled it.
        """
        logging.info("Received kill request")
        self._node_running_status = NodeRunningStatus.CANCELED
        self.job_complete_queue.put(NodeRunningStatus.CANCELED)

    def init_executor(self):
        """Initialize environment for the executor"""
        pool_size = 1
        self.worker_pool = multiprocessing.pool.ThreadPool(
            pool_size, worker_main, (
                self.job_run_queue,
                self.job_complete_queue,
            )
        )

    def clean_up_executor(self):
        """Clean up the environment created by executor"""
        if self.worker_pool:
            self.worker_pool.terminate()
        self.worker_pool = None

    def _apply_inputs(self, node):
        for node_input in node.inputs:
            for input_reference in node_input.input_references:
                node_input.values.extend(
                    self.node_id_to_node[to_object_id(input_reference.node_id)].get_output_by_name(
                        input_reference.output_id
                    ).values
                )

    def run(self, preview: bool = False) -> str:
        """Main execution function.
        """
        assert self.node, "Attribute `node` is unassigned"
        if preview:
            raise Exception("`preview` is not supported for the DAG")

        for sub_node in node_utils.traverse_in_order(self.node):
            if NodeRunningStatus.is_finished(sub_node.node_running_status):
                continue
            self._apply_inputs(sub_node)
            sub_node.node_running_status = NodeRunningStatus.RUNNING
            self.node.save(collection=Collections.RUNS, force=True)

            # Run
            self.job_run_queue.put(sub_node.copy())
            new_node = self.job_complete_queue.get()

            # In case something else is on the queue (i.e. canceled)
            if isinstance(new_node, Node):
                self.update_node(new_node)

            if NodeRunningStatus.is_finished(self._node_running_status):
                prev_status = self._node_running_status
                self._node_running_status = prev_status
                break
            self.node.save(collection=Collections.RUNS, force=True)

        if self._node_running_status == NodeRunningStatus.FAILED_WAITING:
            self._node_running_status = NodeRunningStatus.FAILED
        return self._node_running_status



"""Common utils of the web service."""

import logging
import re
import traceback
from functools import wraps
from typing import Any, Dict, Optional

from flask import Flask, Response, g, request
from flask.logging import create_logger
from flask_cors import CORS

from plynx.constants import RegisterUserExceptionCode, ResponseStatus
from plynx.db.user import User, UserCollectionManager
from plynx.utils.common import JSONEncoder
from plynx.utils.config import get_config
from plynx.utils.content import create_default_templates
from plynx.utils.db_connector import check_connection
from plynx.utils.exceptions import RegisterUserException

app = Flask(__name__)
logger = create_logger(app)

DEFAULT_EMAIL = ''
DEFAULT_USERNAME = 'default'
DEFAULT_PASSWORD = ''

_CONFIG = None


def verify_password(username_or_token: str, password: str):
    """Veryfy password based on user"""
    if _CONFIG and _CONFIG.auth.secret_key and username_or_token == DEFAULT_USERNAME:
        return False
    user = User.verify_auth_token(username_or_token)
    if not user:
        # try to authenticate with username/password
        user = UserCollectionManager.find_user_by_name(username_or_token)
        if not user or not user.verify_password(password):
            return False
    g.user = user   # pylint: disable=assigning-non-slot
    return True


def authenticate():
    """Return 401 message"""
    # 401 response
    return Response(
        'Could not verify your access level for that URL; You have to login with proper credentials',
        401,
        {'WWW-Authenticate': 'PlynxBasic realm="Login Required"'}
    )


def requires_auth(f):
    """Auth wrapper"""
    @wraps(f)
    def decorated(*args, **kwargs):
        auth = request.authorization
        if not auth:
            # if auth not provided, try default
            auth = {'username': DEFAULT_USERNAME, 'password': DEFAULT_PASSWORD}
        if not verify_password(auth['username'], auth['password']):
            return authenticate()
        return f(*args, **kwargs)
    return decorated


def register_user(username: str, password: str, email: str):
    """Register a new user.

    Args:
        username    (str):  Username
        password    (str):  Pasword
        email       (str):  Email

    Return:
        (User):     New user DB Object
    """
    if not username:
        raise RegisterUserException(
            'Missing username',
            error_code=RegisterUserExceptionCode.EMPTY_USERNAME
        )
    if username != DEFAULT_USERNAME and not password:
        raise RegisterUserException(
            'Missing password',
            error_code=RegisterUserExceptionCode.EMPTY_PASSWORD
        )
    if UserCollectionManager.find_user_by_name(username):
        raise RegisterUserException(
            'Username is taken',
            error_code=RegisterUserExceptionCode.USERNAME_ALREADY_EXISTS
        )
    if username != DEFAULT_USERNAME and not re.match(r"(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)", email):
        raise RegisterUserException(
            f"Invalid email: `{email}`",
            error_code=RegisterUserExceptionCode.INVALID_EMAIL
        )
    if username != DEFAULT_USERNAME and UserCollectionManager.find_user_by_email(email):
        raise RegisterUserException(
            'Email already exists',
            error_code=RegisterUserExceptionCode.EMAIL_ALREADY_EXISTS
        )
    if len(username) < 6 or len(username) > 22:
        raise RegisterUserException(
            'Lenght of the username must be between 6 and 22',
            error_code=RegisterUserExceptionCode.INVALID_LENGTH_OF_USERNAME
        )

    user = User()
    user.username = username
    user.email = email
    user.hash_password(password)
    user.save()
    return user


def _init_default_user():
    if UserCollectionManager.find_user_by_name(DEFAULT_USERNAME) is None:
        user = register_user(DEFAULT_USERNAME, DEFAULT_PASSWORD, DEFAULT_EMAIL)
        logging.info(f"Created default user `{DEFAULT_USERNAME}`")
        create_default_templates(user)
    else:
        logging.info(f"Default user `{DEFAULT_USERNAME}` already exists")


def make_fail_response(message, **kwargs):
    """Return basic fail response"""
    return JSONEncoder().encode({
        'status': ResponseStatus.FAILED,
        'message': message,
        **kwargs
    })


def make_permission_denied(message: str = 'Permission denied'):
    """Return permission error"""
    return make_fail_response(message), 403


def make_success_response(extra_response: Optional[Dict[str, Any]] = None):
    """Return successful response"""
    return JSONEncoder().encode(dict(
        {
            'status': ResponseStatus.SUCCESS,
            'message': 'Success',
            'settings': g.user.settings.to_dict(),
        }, **(extra_response or {})))


def handle_errors(f):
    """Handle errors wrapper"""
    @wraps(f)
    def decorated(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except Exception as e:  # pylint: disable=broad-except
            logger.error(traceback.format_exc())
            return make_fail_response(f'Internal error: "{repr(e)}"'), 500
    return decorated


def run_api():
    """Run web service"""
    global _CONFIG  # pylint: disable=global-statement
    _CONFIG = get_config()

    check_connection()

    # Create default user if in single user mode
    if not _CONFIG.auth.secret_key:
        logging.info('Single user mode')
        _init_default_user()
    else:
        logging.info('Multiple user mode')

    CORS(app, resources={r"/*": {"origins": "*"}})
    app.run(
        host=_CONFIG.web.host,
        port=_CONFIG.web.port,
        debug=_CONFIG.web.debug,
        use_reloader=False,
    )



"""User DB Object and utils"""

from dataclasses import dataclass, field
from datetime import datetime, timedelta, timezone
from typing import Dict, List, Optional, Union

import jwt
from dataclasses_json import dataclass_json
from passlib.apps import custom_app_context as pwd_context

from plynx.constants import Collections, TokenType
from plynx.db.db_object import DBObject
from plynx.utils.common import ObjectId
from plynx.utils.config import get_auth_config, get_iam_policies_config
from plynx.utils.db_connector import get_db_connector

DEFAULT_POLICIES = get_iam_policies_config().default_policies
JWT_ENCODE_ALGORITHM = "HS256"


@dataclass_json
@dataclass
class UserSettings(DBObject):
    """User Settings structure."""
    display_name: str = ""


@dataclass_json
@dataclass
class User(DBObject):
    """Basic User class with db interface."""
    DB_COLLECTION = Collections.USERS

    _id: ObjectId = field(default_factory=ObjectId)
    username: str = ""
    email: Optional[str] = ""
    password_hash: str = ""
    active: bool = True
    policies: List[str] = field(default_factory=lambda: list(DEFAULT_POLICIES))
    settings: UserSettings = field(default_factory=UserSettings)

    def hash_password(self, password: str):
        """Change password.

        Args:
            password    (str)   Real password string
        """
        self.password_hash = pwd_context.encrypt(password)  # pylint: disable=attribute-defined-outside-init

    def verify_password(self, password: str) -> bool:
        """Verify password.

        Args:
            password    (str)   Real password string

        Return:
            (bool)    True if password matches else False
        """
        return pwd_context.verify(password, self.password_hash)

    def check_role(self, role: str) -> bool:
        """Check if the user has a given role"""
        return role in self.policies

    @staticmethod
    def find_users() -> List[Dict]:
        """Get all the users"""
        return getattr(get_db_connector(), User.DB_COLLECTION).find({})

    def generate_token(self, token_type: str, expiration: int = 600) -> str:
        """Generate a token.

        Args:
            token_type  (str)   Either TokenType.ACCESS_TOKEN, or TokenType.REFRESH_TOKEN
            expiration  (int)   Time to Live (TTL) in sec

        Return:
            (str)   Secured token
        """
        payload: Dict[str, Union[str, datetime]] = {
            "username": self.username,
            "type": token_type,
        }

        if token_type == TokenType.ACCESS_TOKEN:
            payload["exp"] = datetime.now(tz=timezone.utc) + timedelta(seconds=expiration)
        elif token_type == TokenType.REFRESH_TOKEN:
            payload["exp"] = datetime.now(tz=timezone.utc) + timedelta(hours=720)
        else:
            raise ValueError(f"`token_type` is unknown value `{token_type}`")

        token = jwt.encode(payload, get_auth_config().secret_key, algorithm=JWT_ENCODE_ALGORITHM)

        return token

    @staticmethod
    def verify_auth_token(token: str) -> Optional["User"]:
        """Verify token.

        Args:
            token   (str)   Token

        Return:
            (User)   User object or None
        """
        try:
            payload = jwt.decode(token, get_auth_config().secret_key, algorithms=[JWT_ENCODE_ALGORITHM])
        except (jwt.ExpiredSignatureError, jwt.InvalidTokenError):
            return None
        except Exception as e:  # pylint: disable=broad-except
            print(f"Unexpected exception: {e}")
            return None

        user = UserCollectionManager.find_user_by_name(payload['username'])
        if not user or not user.active:
            return None
        return user


class UserCollectionManager:
    """User Manger"""
    @staticmethod
    def find_user_by_name(username: str) -> Optional[User]:
        """Find User.

        Args:
            username    (str)   Username

        Return:
            (User)   User object or None
        """
        user_dict = getattr(get_db_connector(), User.DB_COLLECTION).find_one({'username': username})
        if not user_dict:
            return None

        return User.from_dict(user_dict)

    @staticmethod
    def find_user_by_email(email: str) -> Optional[User]:
        """Find User.

        Args:
            email    (str)   Email

        Return:
            (User)   User object or None
        """
        user_dict = getattr(get_db_connector(), User.DB_COLLECTION).find_one({'email': email})
        if not user_dict:
            return None

        return User(user_dict)

    @staticmethod
    def get_users(
            search: str = "",
            per_page: int = 20,
            offset: int = 0,
            ):
        """Get a list of users"""
        raise NotImplementedError()



"""Utils that materialize plugins"""
# pylint: disable=global-statement
import logging
import pydoc
from typing import Any, Dict

from plynx.base.executor import BaseExecutor
from plynx.base.hub import BaseHub
from plynx.base.resource import BaseResource
from plynx.utils.config import PluginsConfig, get_plugins


def _isinstance_namedtuple(x: Any) -> bool:
    return isinstance(x, tuple) and getattr(x, '_fields', None) is not None


def _as_dict(obj: Any):
    if _isinstance_namedtuple(obj):
        return {
            key: _as_dict(getattr(obj, key)) for key in obj._fields
        }
    elif isinstance(obj, list):
        return [_as_dict(value) for value in obj]
    return obj


class _ResourceManager:
    def __init__(self, plugins: PluginsConfig):
        self.resources = plugins.resources
        self.kind_to_resource_class: Dict[str, BaseResource] = {
            resource.kind: pydoc.locate(resource.cls)   # type: ignore
            for resource in self.resources
        }
        self.kind_to_resource_dict: Dict[str, Dict[str, Any]] = {
            resource.kind: resource._asdict() for resource in self.resources
        }
        for kind, resource_class in self.kind_to_resource_class.items():
            self.kind_to_resource_dict[kind]['display_raw'] = resource_class.DISPLAY_RAW


class _ExecutorManager:
    def __init__(self, plugins: PluginsConfig):
        self.kind_to_executor_class: Dict[str, BaseExecutor] = {}
        self.kind_to_icon = {}
        self.kind_to_color = {}
        self.kind_to_title = {}
        for o_or_w in plugins.workflows + plugins.operations + plugins.dummy_operations:
            self.kind_to_executor_class[o_or_w.kind] = pydoc.locate(o_or_w.executor)    # type: ignore
            if not self.kind_to_executor_class[o_or_w.kind]:
                raise Exception(f"Executor `{o_or_w.executor}` not found")
            self.kind_to_icon[o_or_w.kind] = o_or_w.icon
            self.kind_to_color[o_or_w.kind] = o_or_w.color
            self.kind_to_title[o_or_w.kind] = o_or_w.title

        self.kind_info = {}
        for kind, executor_class in self.kind_to_executor_class.items():
            logging.info(f"Initializing executor `{kind}`")
            self.kind_info[kind] = {
                'is_graph': executor_class.IS_GRAPH,
                'title': self.kind_to_title[kind],
                'icon': self.kind_to_icon[kind],
                'color': self.kind_to_color[kind],
            }


class _OperationManager:
    def __init__(self, plugins: PluginsConfig):
        self.operations = plugins.operations
        self.kind_to_operation_dict = {
            operation.kind: _as_dict(operation) for operation in self.operations
        }


class _HubManager:
    def __init__(self, plugins: PluginsConfig):
        self.hubs = plugins.hubs
        self.kind_to_hub_class: Dict[str, BaseHub] = {
            hub.kind: pydoc.locate(hub.cls)(**hub.args)     # type: ignore
            for hub in self.hubs
        }
        self.kind_to_hub_dict = {
            hub.kind: hub._asdict() for hub in self.hubs
        }


class _WorkflowManager:
    def __init__(self, plugins: PluginsConfig):
        self.workflows = plugins.workflows
        self.kind_to_workflow_dict = {
            workflow.kind: workflow._asdict() for workflow in self.workflows
        }


_plugins: PluginsConfig = get_plugins()

_RESOURCE_MANAGER = None
_OPERATION_MANAGER = None
_HUB_MANAGER = None
_WORKFLOW_MANAGER = None
_EXECUTOR_MANAGER = None

_PLUGINS_DICT = None


def get_resource_manager():
    """Generate Resource plugin structure"""
    global _RESOURCE_MANAGER
    if not _RESOURCE_MANAGER:
        _RESOURCE_MANAGER = _ResourceManager(_plugins)
    return _RESOURCE_MANAGER


def get_operation_manager():
    """Generate Operation plugin structure"""
    global _OPERATION_MANAGER
    if not _OPERATION_MANAGER:
        _OPERATION_MANAGER = _OperationManager(_plugins)
    return _OPERATION_MANAGER


def get_hub_manager():
    """Generate Hub plugin structure"""
    global _HUB_MANAGER
    if not _HUB_MANAGER:
        _HUB_MANAGER = _HubManager(_plugins)
    return _HUB_MANAGER


def get_workflow_manager():
    """Generate Workflow plugin structure"""
    global _WORKFLOW_MANAGER
    if not _WORKFLOW_MANAGER:
        _WORKFLOW_MANAGER = _WorkflowManager(_plugins)
    return _WORKFLOW_MANAGER


def get_executor_manager():
    """Generate Exectutor plugin structure"""
    global _EXECUTOR_MANAGER
    if not _EXECUTOR_MANAGER:
        _EXECUTOR_MANAGER = _ExecutorManager(_plugins)
    return _EXECUTOR_MANAGER


def get_plugins_dict():
    """Generate all of the plugins structure"""
    global _PLUGINS_DICT
    if not _PLUGINS_DICT:
        _PLUGINS_DICT = {
            'resources_dict': get_resource_manager().kind_to_resource_dict,
            'operations_dict': get_operation_manager().kind_to_operation_dict,
            'hubs_dict': get_hub_manager().kind_to_hub_dict,
            'workflows_dict': get_workflow_manager().kind_to_workflow_dict,
            'executors_info': get_executor_manager().kind_info,
        }
    return _PLUGINS_DICT



"""This module contais utils related to plynx.db.Node, but not necessary involved into DB structure"""
import hashlib
import logging as logger
from collections import defaultdict, deque
from typing import Deque, Dict, List, Optional, Set, Tuple

import plynx.db.node_collection_manager
from plynx.constants import IGNORED_CACHE_PARAMETERS, Collections, NodeClonePolicy, NodeRunningStatus, ParameterTypes, SpecialNodeId
from plynx.db.node import CachedNode, Node
from plynx.utils.common import ObjectId, to_object_id

node_collection_managers = {
    collection: plynx.db.node_collection_manager.NodeCollectionManager(collection=collection)
    for collection in [Collections.TEMPLATES, Collections.RUNS]
}


class _GraphVertex:
    """Used for internal purposes."""

    def __init__(self):
        self.edges = []
        self.num_connections = 0


class GraphError(Exception):
    """Generic Graph topology exception"""


def _generate_parameters_key(node: Node) -> str:
    """Generate hash key based on parameters only.

    Args:
        node    (Node): Node object

    Return:
        (str)   Hash value
    """

    parameters = node.parameters

    sorted_parameters = sorted(parameters, key=lambda x: x.name)
    parameters_hash = ','.join([
        f"{parameter.name}:{parameter.value}"
        for parameter in sorted_parameters if parameter.name not in IGNORED_CACHE_PARAMETERS
    ])

    return hashlib.sha256(
        ';'.join([
                parameters_hash,
            ]).encode('utf-8')
    ).hexdigest()


# pylint: disable=too-many-locals
def augment_node_with_cache(node: Node, other_node: Node) -> None:
    """
    Augment the Node in templates with a Node in Run.
    Results will be stored in `_cached_node` fields of the subnodes and not applied directly.
    """
    # pylint: disable=too-many-branches
    # TODO optimize function and remove too-many-locals
    node._cached_node = None
    sub_nodes_parameter = node.get_parameter_by_name('_nodes', throw=False)
    if not sub_nodes_parameter:
        # TODO check if cacheable
        # TODO probably never called.
        # TODO Update when run augmentation recursevely
        raise NotImplementedError("Subnodes are not found")

    sub_nodes = sub_nodes_parameter.value.value
    other_sub_nodes = other_node.get_parameter_by_name('_nodes').value.value
    other_node_id_to_original_id = {}
    for other_sub_node in other_sub_nodes:
        other_node_id_to_original_id[other_sub_node._id] = other_sub_node.template_node_id

    id_to_node = {}
    for sub_node in sub_nodes:
        sub_node._cached_node = None
        obj_id = to_object_id(sub_node._id)
        id_to_node[obj_id] = sub_node

    for other_subnode in traverse_in_order(other_node):
        if other_subnode.template_node_id not in id_to_node:
            continue
        sub_node = id_to_node[other_subnode.template_node_id]

        # TODO: check is final state
        this_cache = _generate_parameters_key(sub_node)
        other_node_cache = _generate_parameters_key(other_subnode)

        if this_cache != other_node_cache:
            continue

        tmp_inputs = []
        for input in sub_node.inputs:   # pylint: disable=redefined-builtin
            for input_reference in input.input_references:
                tmp_inputs.append(f"{input_reference.node_id}-{input_reference.output_id}")
        sub_node_inputs_hash = ",".join(tmp_inputs)

        tmp_inputs = []
        for input in other_subnode.inputs:  # pylint: disable=redefined-builtin
            for input_reference in input.input_references:
                orig_idx = other_node_id_to_original_id.get(to_object_id(input_reference.node_id), 'none')
                tmp_inputs.append(f"{orig_idx}-{input_reference.output_id}")
        other_subnode_inputs_hash = ",".join(tmp_inputs)

        if sub_node_inputs_hash != other_subnode_inputs_hash:
            continue

        tmp_refs_is_cached = False
        for input in sub_node.inputs:
            for input_reference in input.input_references:
                ref_node_id = input_reference.node_id
                if id_to_node[to_object_id(ref_node_id)]._cached_node is None:
                    tmp_refs_is_cached = True
        if tmp_refs_is_cached:
            continue

        sub_node._cached_node = CachedNode(
            node_running_status=other_subnode.node_running_status,
            outputs=other_subnode.outputs,
            logs=other_subnode.logs,
        )


def traverse_reversed(node: Node):
    """
    Traverse the subnodes in a reversed from the topoligical order.
    """
    # pylint: disable=too-many-branches
    sub_nodes_parameter = node.get_parameter_by_name('_nodes', throw=False)
    if not sub_nodes_parameter:
        yield node
        return

    sub_nodes = sub_nodes_parameter.value.value
    if len(sub_nodes) == 0:
        return

    id_to_vertex = {sub_node._id: _GraphVertex() for sub_node in sub_nodes}
    dfs_queue: Deque[ObjectId] = deque()
    node_index = {}

    for sub_node in sub_nodes:
        node_index[sub_node._id] = sub_node
        for input in sub_node.inputs:   # pylint: disable=redefined-builtin
            for input_reference in input.input_references:
                ref_node_id = to_object_id(input_reference.node_id)
                id_to_vertex[sub_node._id].edges.append(ref_node_id)
                id_to_vertex[ref_node_id].num_connections += 1

    for vertex_id, vertex in id_to_vertex.items():
        if vertex.num_connections == 0:
            dfs_queue.append(vertex_id)

    if len(dfs_queue) == 0:
        raise GraphError("No node without outgoing output found")

    while dfs_queue:
        node_id = dfs_queue.popleft()
        yield node_index[node_id]
        for vertex_id in id_to_vertex[node_id].edges:
            id_to_vertex[vertex_id].num_connections -= 1
            if id_to_vertex[vertex_id].num_connections == 0:
                dfs_queue.append(vertex_id)

    for vertex_id, vertex in id_to_vertex.items():
        if vertex.num_connections != 0:
            raise GraphError("Unresolved connections")


def traverse_in_order(node: Node):
    """
    Traverse the subnodes in a topoligical order.
    """
    nodes = list(traverse_reversed(node))
    return reversed(nodes)


# pylint: disable=inconsistent-return-statements
def arrange_auto_layout(node: Node, readonly: bool = False):
    """Use heuristic to rearange nodes."""
    # pylint: disable=invalid-name,too-many-branches,too-many-locals,too-many-statements
    HEADER_HEIGHT = 23
    TITLE_HEIGHT = 20
    FOOTER_HEIGHT = 10
    BORDERS_HEIGHT = 2
    # ITEM_HEIGHT = 20
    ITEM_HEIGHT = 30
    OUTPUT_ITEM_HEIGHT = 100
    SPACE_HEIGHT = 50
    LEFT_PADDING = 30
    TOP_PADDING = 80
    LEVEL_WIDTH = 252
    SPECIAL_PARAMETER_HEIGHT = 20
    SPECIAL_PARAMETER_TYPES = [ParameterTypes.CODE]
    min_node_height = HEADER_HEIGHT + TITLE_HEIGHT + FOOTER_HEIGHT + BORDERS_HEIGHT

    node_id_to_level = defaultdict(lambda: -1)
    node_id_to_node = {}
    queued_node_ids = set()
    children_ids = defaultdict(set)

    sub_nodes = node.get_parameter_by_name('_nodes').value.value

    if len(sub_nodes) == 0:
        return

    node_ids = {node._id for node in sub_nodes}
    non_zero_node_ids = set()
    for sub_node in sub_nodes:
        node_id_to_node[node._id] = sub_node
        for input in sub_node.inputs:   # pylint: disable=redefined-builtin
            for input_reference in input.input_references:
                parent_node_id = ObjectId(input_reference.node_id)
                non_zero_node_ids.add(parent_node_id)
                children_ids[parent_node_id].add(sub_node._id)

    leaves = node_ids - non_zero_node_ids
    to_visit: deque = deque()
    # Alwasy put Output Node in the end
    push_special = SpecialNodeId.OUTPUT in leaves and len(leaves) > 1
    for leaf_id in leaves:
        node_id_to_level[leaf_id] = 1 if push_special and leaf_id != SpecialNodeId.OUTPUT else 0
        to_visit.append(leaf_id)

    while to_visit:
        node_id = to_visit.popleft()
        sub_node = node_id_to_node[node_id]
        node_level = max([node_id_to_level[node_id]] + [node_id_to_level[child_id] + 1 for child_id in children_ids[node_id]])
        node_id_to_level[node_id] = node_level
        for input in node.inputs:   # pylint: disable=redefined-builtin
            for input_reference in input.input_references:
                parent_node_id = ObjectId(input_reference.node_id)
                parent_level = node_id_to_level[parent_node_id]
                node_id_to_level[parent_node_id] = max(node_level + 1, parent_level)
                if parent_node_id not in queued_node_ids:
                    to_visit.append(parent_node_id)
                    queued_node_ids.add(parent_node_id)

    max_level = max(node_id_to_level.values())
    level_to_node_ids: Dict[int, List[ObjectId]] = defaultdict(list)
    row_heights: Dict[int, int] = defaultdict(lambda: 0)

    def get_index_helper(node, level):
        if level < 0:
            return 0
        parent_node_ids = set()
        for input in node.inputs:   # pylint: disable=redefined-builtin
            for input_reference in input.input_references:
                parent_node_ids.add(ObjectId(input_reference.node_id))

        for index, node_id in enumerate(level_to_node_ids[level]):
            if node_id in parent_node_ids:
                return index
        return -1

    def get_index(node, max_level, level):
        # pylint: disable=consider-using-generator
        return tuple(
            [get_index_helper(node, lvl) for lvl in range(max_level, level, -1)]
        )

    for node_id, level in node_id_to_level.items():
        level_to_node_ids[level].append(node_id)

    # Push Input Node up the level
    if SpecialNodeId.INPUT in node_id_to_level and \
            (node_id_to_level[SpecialNodeId.INPUT] != max_level or len(level_to_node_ids[max_level]) > 1):
        input_level = node_id_to_level[SpecialNodeId.INPUT]
        level_to_node_ids[input_level] = [node_id for node_id in level_to_node_ids[input_level] if node_id != SpecialNodeId.INPUT]
        max_level += 1
        node_id_to_level[SpecialNodeId.INPUT] = max_level
        level_to_node_ids[max_level] = [SpecialNodeId.INPUT]

    for level in range(max_level, -1, -1):
        level_node_ids = level_to_node_ids[level]
        index_to_node_id = []
        for node_id in level_node_ids:
            node = node_id_to_node[node_id]
            index = get_index(node, max_level, level)
            index_to_node_id.append((index, node_id))

        index_to_node_id.sort()
        level_to_node_ids[level] = [node_id for _, node_id in index_to_node_id]

        for index, node_id in enumerate(level_to_node_ids[level]):
            node = node_id_to_node[node_id]
            special_parameters_count = sum(
                1 if parameter.parameter_type in SPECIAL_PARAMETER_TYPES and parameter.widget else 0
                for parameter in node.parameters
            )
            node_height = sum([
                min_node_height,
                ITEM_HEIGHT * len(node.inputs) + OUTPUT_ITEM_HEIGHT * len(node.outputs),
                special_parameters_count * SPECIAL_PARAMETER_HEIGHT
            ])
            row_heights[index] = max(row_heights[index], node_height)

    # TODO compute grid in a separate function
    if readonly:
        return level_to_node_ids, node_id_to_node

    cum_heights = [0]
    for row_height in row_heights:
        cum_heights.append(cum_heights[-1] + row_height + SPACE_HEIGHT)

    max_height = max(cum_heights)

    for level in range(max_level, -1, -1):
        level_node_ids = level_to_node_ids[level]
        level_height = cum_heights[len(level_node_ids)]
        level_padding = (max_height - level_height) // 2
        for index, node_id in enumerate(level_node_ids):
            node = node_id_to_node[node_id]
            node.x = LEFT_PADDING + (max_level - level) * LEVEL_WIDTH
            node.y = TOP_PADDING + level_padding + cum_heights[index]


def reset_failed_nodes(node: Node):
    """
    Reset the statuses of the failed subnodes.

    Resetting means node_running_status as well as the outputs and the logs.
    """
    sub_nodes_parameter = node.get_parameter_by_name('_nodes', throw=False)
    if not sub_nodes_parameter:
        return

    sub_nodes = sub_nodes_parameter.value.value
    for sub_node in sub_nodes:
        if not NodeRunningStatus.is_failed(sub_node.node_running_status):
            continue
        sub_node.node_running_status = NodeRunningStatus.READY
        for resource in sub_node.outputs + sub_node.logs:
            resource.values = []


def apply_cache(node: Node):
    """Apply cache values to outputs and logs"""
    sub_nodes_parameter = node.get_parameter_by_name('_nodes', throw=False)
    if not sub_nodes_parameter:
        # TODO check if cacheable
        raise NotImplementedError("Subnodes not found. Do we want to be here?")

    sub_nodes = sub_nodes_parameter.value.value

    for sub_node in sub_nodes:
        if not sub_node._cached_node:
            continue
        sub_node.node_running_status = sub_node._cached_node.node_running_status
        sub_node.outputs = sub_node._cached_node.outputs
        sub_node.logs = sub_node._cached_node.logs

        sub_node._cached_node = None


def is_all_cached_and_successful(node: Node):
    """
    Check if the Node in a run needs recomputing at all.
    """
    sub_nodes_parameter = node.get_parameter_by_name('_nodes', throw=False)
    if not sub_nodes_parameter:
        # TODO check if cacheable
        return True

    sub_nodes = sub_nodes_parameter.value.value

    for sub_node in sub_nodes:
        if not sub_node._cached_node:
            return False
        if NodeRunningStatus.is_failed(sub_node._cached_node.node_running_status):
            return False
    return True


def construct_new_run(node: Node, user_id) -> Tuple[Optional[Node], Node]:
    """
    Create a new run based on a Node itself and the latest run as well.
    """
    node = Node.from_dict(node.to_dict())
    node_in_run: Optional[Node] = None
    if node.latest_run_id:
        node_in_run_dict = node_collection_managers[Collections.RUNS].get_db_node(node.latest_run_id, user_id)

        if node_in_run_dict:
            node_in_run = Node.from_dict(
                node_in_run_dict
            )
            augment_node_with_cache(node, node_in_run)
            apply_cache(node)
        else:
            logger.warning(f"Failed to load a run with id `{node.latest_run_id}`")

    new_node_in_run = node.clone(NodeClonePolicy.NODE_TO_RUN, override_finished_state=False)
    reset_failed_nodes(new_node_in_run)

    return node_in_run, new_node_in_run


def remove_auto_run_disabled(node: Node):
    """
    Trim the subnodes the way that if there is no need to run a subnode and it is not auto runnable, ignore it.
    """
    node_ids_to_remove = set()
    for sub_node in traverse_in_order(node):
        if (not sub_node.auto_run or not sub_node.auto_run_enabled) and not NodeRunningStatus.is_succeeded(sub_node.node_running_status):
            node_ids_to_remove.add(sub_node._id)
        for input in sub_node.inputs:   # pylint: disable=redefined-builtin
            for input_reference in input.input_references:
                if input_reference.node_id in node_ids_to_remove:
                    node_ids_to_remove.add(sub_node._id)

    if len(node_ids_to_remove) == 0:
        return

    sub_nodes_parameter = node.get_parameter_by_name('_nodes', throw=False)
    if not sub_nodes_parameter:
        logger.warning("remove_auto_run_disabled(..): are we supposed to be here?")
        return
    sub_nodes = sub_nodes_parameter.value.value

    sub_nodes_parameter.value.value = list(filter(lambda sn: sn._id not in node_ids_to_remove, sub_nodes))


def calc_status_to_node_ids(node: Optional[Node]) -> Dict[str, Set[ObjectId]]:
    """
    Make a map node_running_status to list of ids.
    """
    res: Dict[str, Set[ObjectId]] = defaultdict(set)
    if not node:
        return res
    for sub_node in node.get_sub_nodes():
        res[sub_node.node_running_status].add(sub_node._id)
    return res



"""A standard executor for DAGs."""
import functools
import logging
import time
from collections import defaultdict
from typing import Dict, List, Optional

import plynx.base.executor
import plynx.db.node_cache_manager
import plynx.plugins.executors.bases
import plynx.utils.executor
from plynx.constants import Collections, NodeRunningStatus, ParameterTypes, SpecialNodeId, ValidationCode, ValidationTargetType
from plynx.db.node import Node, Parameter
from plynx.db.validation_error import ValidationError
from plynx.utils.common import ObjectId, to_object_id

_WAIT_STATUS_BEFORE_FAILED = {
    NodeRunningStatus.RUNNING,
    NodeRunningStatus.IN_QUEUE,
    NodeRunningStatus.FAILED_WAITING,
}
_ACTIVE_WAITING_TO_STOP = {
    NodeRunningStatus.FAILED_WAITING,
    NodeRunningStatus.CANCELED,
}


@functools.lru_cache()
def node_cache_manager():
    """Lazy NodeCacheManager definition"""
    return plynx.db.node_cache_manager.NodeCacheManager()


class DAG(plynx.plugins.executors.bases.PLynxAsyncExecutor):
    """ Main graph scheduler.

    Args:
        node (Node)

    """
    # pylint: disable=too-many-instance-attributes

    IS_GRAPH = True
    GRAPH_ITERATION_SLEEP = 1

    def __init__(self, node: Node):
        # pylint: disable=too-many-branches
        super().__init__(node)
        assert self.node, "Attribute `node` is not defined"

        self.subnodes: List[Node] = self.node.get_parameter_by_name('_nodes', throw=True).value.value

        self.node_id_to_node: Dict[ObjectId, Node] = {
            node._id: node for node in self.subnodes
        }

        # number of dependencies to ids
        self.dependency_index_to_node_ids = defaultdict(set)
        self.node_id_to_dependents = defaultdict(set)
        self.node_id_to_dependency_index = defaultdict(lambda: 0)
        self.uncompleted_nodes_count = 0

        self._node_running_status = NodeRunningStatus.READY

        for subnode in self.subnodes:
            node_id = subnode._id
            if node_id == SpecialNodeId.INPUT:
                updated_resources_count = 0
                for output in subnode.outputs:
                    for input in self.node.inputs:  # pylint: disable=redefined-builtin
                        if input.name == output.name:
                            updated_resources_count += 1
                            output.values = input.values
                if updated_resources_count != len(self.node.inputs):
                    raise Exception(f"Used {updated_resources_count} inputs for {len(self.node.inputs)} outputs")

            # ignore nodes in finished statuses
            if NodeRunningStatus.is_finished(subnode.node_running_status) and node_id != SpecialNodeId.OUTPUT:
                continue
            dependency_index = 0
            for node_input in subnode.inputs:
                for input_reference in node_input.input_references:
                    dep_node_id = to_object_id(input_reference.node_id)
                    self.node_id_to_dependents[dep_node_id].add(node_id)
                    if not NodeRunningStatus.is_finished(self.node_id_to_node[dep_node_id].node_running_status):
                        dependency_index += 1

            if not NodeRunningStatus.is_finished(subnode.node_running_status):
                self.uncompleted_nodes_count += 1
            self.dependency_index_to_node_ids[dependency_index].add(node_id)
            self.node_id_to_dependency_index[node_id] = dependency_index

        self.monitoring_executors: List[plynx.base.executor.BaseExecutor] = []

        if self.uncompleted_nodes_count == 0:
            self._node_running_status = NodeRunningStatus.SUCCESS

    def finished(self) -> bool:
        """Return True or False depending on the running status of the DAG."""
        if self._node_running_status in _ACTIVE_WAITING_TO_STOP:
            # wait for the rest of the running jobs to finish
            # check running status of each of the nodes
            for node in self.subnodes:
                if node.node_running_status in _WAIT_STATUS_BEFORE_FAILED:
                    return False

            # set status to FAILED
            if self._node_running_status == NodeRunningStatus.FAILED_WAITING:
                self._node_running_status = NodeRunningStatus.FAILED
            return True
        return self._node_running_status in {NodeRunningStatus.SUCCESS, NodeRunningStatus.FAILED, NodeRunningStatus.CANCELED}

    def pop_jobs(self) -> List[Node]:
        """Get a set of nodes with satisfied dependencies"""
        res: List[Node] = []
        logging.info("Pop jobs")

        finished_node_ids = set()
        for executor in self.monitoring_executors:
            assert executor.node, "executor node must be defined at this point"
            running_status = executor.get_running_status()

            # check status
            if NodeRunningStatus.is_finished(running_status.node_running_status):
                self.update_node(executor.node)
                finished_node_ids.add(executor.node._id)
        self.monitoring_executors = [ex for ex in self.monitoring_executors if ex.node._id not in finished_node_ids]    # type: ignore

        if NodeRunningStatus.is_failed(self._node_running_status):
            logging.info("Job in DAG failed, pop_jobs will return []")
            return res

        cached_nodes = []
        for node_id in self.dependency_index_to_node_ids[0]:
            # Get the node and init its inputs, i.e. filling its resource_ids
            orig_node = self.node_id_to_node[node_id]
            for node_input in orig_node.inputs:
                for input_reference in node_input.input_references:
                    node_input.values.extend(
                        self.node_id_to_node[to_object_id(input_reference.node_id)].get_output_by_name(
                            input_reference.output_id
                        ).values
                    )
            orig_node.node_running_status = NodeRunningStatus.IN_QUEUE
            node = orig_node.copy()   # type: ignore

            if DAG._cacheable(node):
                try:
                    cache = node_cache_manager().get(node)
                    if cache:
                        node.node_running_status = NodeRunningStatus.RESTORED
                        node.outputs = cache.outputs
                        node.logs = cache.logs
                        node.cache_url = f"/runs/{cache.run_id}?nid={cache.node_id}"
                        cached_nodes.append(node)
                        continue
                except Exception as err:    # pylint: disable=broad-except
                    logging.exception(f"Unable to update cache: `{err}`")
            res.append(node)
        del self.dependency_index_to_node_ids[0]

        for node in cached_nodes:
            self.update_node(node)
        if cached_nodes:
            node.save(collection=Collections.RUNS)

        return res

    def update_node(self, node: Node):
        """
        Update node_running_status and outputs if the state has changed.
        """
        assert self.node, "Attribute `node` is unassigned"
        dest_node = self.node_id_to_node[node._id]
        if node.node_running_status == NodeRunningStatus.SUCCESS \
                and dest_node.node_running_status != node.node_running_status \
                and DAG._cacheable(node):
            node_cache_manager().post(node, self.node._id)

        if dest_node.node_running_status == node.node_running_status:
            return

        self._set_node_status(node._id, node.node_running_status)
        # TODO smarter copy
        dest_node.parameters = node.parameters
        dest_node.logs = node.logs
        dest_node.outputs = node.outputs
        dest_node.cache_url = node.cache_url

    def _set_node_status(self, node_id: ObjectId, node_running_status: str):
        node = self.node_id_to_node[node_id]
        node.node_running_status = node_running_status

        logging.info(f"Node running status {node_running_status} {node.title}")

        if node_running_status == NodeRunningStatus.FAILED:
            # TODO optional cancel based on parameter
            self.kill()
            self._node_running_status = NodeRunningStatus.FAILED_WAITING

        if node_running_status in {NodeRunningStatus.SUCCESS, NodeRunningStatus.RESTORED}:
            for dependent_node_id in self.node_id_to_dependents[node_id]:
                dependent_node = self.node_id_to_node[dependent_node_id]
                prev_dependency_index = self.node_id_to_dependency_index[dependent_node_id]

                removed_dependencies = 0
                for node_input in dependent_node.inputs:
                    for input_reference in node_input.input_references:
                        if to_object_id(input_reference.node_id) == to_object_id(node_id):
                            removed_dependencies += 1
                dependency_index = prev_dependency_index - removed_dependencies

                self.dependency_index_to_node_ids[prev_dependency_index].remove(dependent_node_id)
                self.dependency_index_to_node_ids[dependency_index].add(dependent_node_id)
                self.node_id_to_dependency_index[dependent_node_id] = dependency_index
            self.uncompleted_nodes_count -= 1

        if self.uncompleted_nodes_count == 0 and not NodeRunningStatus.is_failed(self._node_running_status):
            self._node_running_status = NodeRunningStatus.SUCCESS

    @staticmethod
    def _cacheable(node: Node) -> bool:
        for parameter in node.parameters:
            if parameter.name == '_cacheable':
                return parameter.value
        return False

    @classmethod
    def get_default_node(cls, is_workflow: bool) -> Node:
        node = super().get_default_node(is_workflow)
        if not is_workflow:
            node.parameters.append(
                Parameter(
                    name="_cacheable",
                    parameter_type=ParameterTypes.BOOL,
                    value=False,
                    mutable_type=False,
                    publicable=False,
                    removable=False,
                )
            )
        node.parameters.append(
            Parameter(
                name="_timeout",
                parameter_type=ParameterTypes.INT,
                value=600,
                mutable_type=False,
                publicable=True,
                removable=False
            )
        )
        node.title = 'New DAG workflow'
        return node

    def _execute_node(self, node: Node):
        assert self.node, "Attribute `node` is unassigned"
        if NodeRunningStatus.is_finished(node.node_running_status):     # NodeRunningStatus.SPECIAL
            return
        node.author = self.node.author                                  # Change it to the author that runs it

        executor = plynx.utils.executor.materialize_executor(node)
        executor.launch()

        self.monitoring_executors.append(executor)

    def run(self, preview: bool = False) -> str:
        assert self.node, "Attribute `node` is unassigned"
        if preview:
            raise Exception("`preview` is not supported for the DAG")
        while not self.finished():
            new_jobs = self.pop_jobs()
            if len(new_jobs) == 0:
                time.sleep(self.GRAPH_ITERATION_SLEEP)
                continue

            for node in new_jobs:
                self._execute_node(node)

        is_succeeded = NodeRunningStatus.is_succeeded(self._node_running_status)
        if is_succeeded:
            for node in self.subnodes:
                if node._id != SpecialNodeId.OUTPUT:
                    continue
                updated_resources_count = 0
                for output in self.node.outputs:
                    for input in node.inputs:   # pylint: disable=redefined-builtin
                        if input.name == output.name:
                            output.values = input.values
                            updated_resources_count += 1
                            break
                if updated_resources_count != len(node.inputs):
                    raise Exception(f"Used {updated_resources_count} inputs for {node.inputs} outputs")
        return self._node_running_status

    def kill(self):
        """Force to kill the process.

        The reason can be the fact it was working too long or parent exectuter canceled it.
        """
        self._node_running_status = NodeRunningStatus.CANCELED
        for executor in self.monitoring_executors:
            executor.kill()

    def validate(self, ignore_inputs: bool = True) -> Optional[ValidationError]:
        assert self.node, "Attribute `node` is unassigned"
        validation_error = super().validate()
        if validation_error:
            return validation_error

        violations = []
        sub_nodes = self.node.get_parameter_by_name('_nodes').value.value

        if len(sub_nodes) == 0:
            violations.append(
                ValidationError(
                    target=ValidationTargetType.GRAPH,
                    object_id=str(self.node._id),
                    validation_code=ValidationCode.EMPTY_GRAPH
                ))

        for node in sub_nodes:
            node_violation = plynx.utils.executor.materialize_executor(node.to_dict()).validate(ignore_inputs=False)
            if node_violation:
                violations.append(node_violation)

        if len(violations) == 0:
            return None

        return ValidationError(
            target=ValidationTargetType.GRAPH,
            object_id=str(self.node._id),
            validation_code=ValidationCode.IN_DEPENDENTS,
            children=violations
        )
